"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging

from panda3d_gemstone.engine import runtime

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.io.file_system import path_exists
from panda3d_gemstone.framework.singleton import Singleton

class SoundSource(InternalObject):
    """
    Represents a sound source in the Gemstone framework
    """

    BAD = 0
    READY = 1
    PLAYING = 2

    def __init__(self):
        InternalObject.__init__(self)
        self.__sound_manager = SoundManager.get_singleton()

    @property
    def source_id(self) -> object:
        """
        Returns this sound source's source id
        """

        return id(self)

    def get_sound(self, name: str) -> object:
        """
        Retrieves a sound from the sound manager by name
        """

        return self.__sound_manager.get_sound(self.source_id, name)

    def sound_exists(self, name: str) -> bool:
        """
        Returns true if the sound name exists in the 
        current sound manager singleton instance
        """

        return self.get_sound(name) != None

    def __has_default_volume(self) -> bool:
        """
        Returns true if the sound source
        has a default volume
        """

        return hasattr(self, 'default_volume')

    def __get_default_volume(self, name: str, default: object = None) -> float:
        """
        Retrieves the sources default volume for a given sound name. Otherwise NoneType
        """

        if not self.__has_default_volume():
            return None

        return self.default_volume.get(name, default)

    def __attempt_set_sound_volume(self, name: str, volume: float) -> None:
        """
        Attempts to set a sonds volume if it exists
        """

        if volume is None:
            return

        sound_obj = self.get_sound(name)
        if sound_obj:
            sound_obj.set_volume(volume)

    def __attempt_manual_update(self, name: str) -> None:
        """
        Attempts to perform a manual sound update if required
        """

        if not runtime.is_interactive():
            return
        
        sound = self.get_sound(name)
        if sound != None:
            sound.update()

    def play_sound(self, name: str) -> None:
        """
        Plays a sound from the current sound source
        """

        volume = self.__get_default_volume(name)
        self.__attempt_set_sound_volume(name, volume)
        self.__sound_manager.play(self.source_id, name)
        self.__attempt_manual_update(name)

    def stop_sound(self, name: str) -> None:
        """
        Stops the requested sound name
        """

        self.__sound_manager.stop(self.source_id, name)
        self.__attempt_manual_update(name)

    def pause_sound(self, name: str) -> None:
        """
        Pauses the requested sound name
        """

        self.__sound_manager.pause(self.source_id, name)
        self.__attempt_manual_update(name)

    def resume_sound(self, name: str) -> None:
        """
        Resumes the requested sound name
        """

        self.__sound_manager.resume(self.source_id, name)
        self.__attempt_manual_update(name)

    def set_sound_loop(self, name: str, flag: bool) -> None:
        """
        Sets the loop flag of the requested sound name
        """

        self.__sound_manager.set_loop(self.source_id, name, flag)
        self.__attempt_manual_update(name)

    def set_sound_loop_count(self, name: str, count: int) -> None:
        """
        Sets the loop count of the requested sound name
        """

        self.__sound_manager.set_loop_count(self.source_id, name, count)
        self.__attempt_manual_update(name)

    def get_sound_status(self, name: str) -> int:
        """
        Returns the sound status of the requested sound name and sound source
        """

        self.__sound_manager.status(self.source_id, name)

    def is_sound_playing(self, name: str) -> bool:
        """
        Returns the sound playing status of the requested sound name from this sound source
        """

        self.get_sound_status(name) == SoundSource.PLAYING

    def is_sound_ready(self, name: str) -> bool:
        """
        Returns the sound ready status of the requested sound name from this sound source
        """

        self.get_sound_status(name) == SoundSource.READY

    def is_sound_bad(self, name: str) -> bool:
        """
        Returns the sound bad status of the requested sound name from this sound source
        """

        self.get_sound_status(name) == SoundSource.BAD

    def add_sound(self, name: str, filename: str) -> None:
        """
        Registers a new sound with the sound manager for this
        sound source
        """

        self.__sound_manager.register(self.source_id, name, filename)

    def add_music(self, name: str, filename: str) -> None:
        """
        Registers a new music with the sound manager for this
        sound source
        """

        self.__sound_manager.register_music(self.source_id, name, filename)

    def remove_sound(self, name: str) -> None:
        """
        Removes the requested sound name from the sound source
        """

        self.__sound_manager.unregister(self.source_id, name)

    def remove_all_sounds(self) -> None:
        """
        Removes all sounds from the sound source
        """

        self.__sound_manager.unregister(self.source_id)

    def copy_sounds(self, new_source: object) -> None:
        """
        Copys all registered sound sources from this sound source to a new source instance
        """

        assert isinstance(new_source, SoundSource)
        self.__sound_manager.copy_sounds(self.source_id, new_source.source_id)

        if self.__has_default_volume():
            if not hasattr(new_source, 'default_volume'):
                new_source.default_volume = {}
            new_source.default_volume.update(self.default_volume)

    def load_sound_data(self, data: dict) -> None:
        """
        Loads sound data from configurable variants
        """

        self.default_volume = {}
        for name, param in list(data.items()):
            if type(param) == type(''):
                self.add_sound(name, param)
            elif type(param) == tuple:
                self.add_sound(name, param[0])

                sound_obj = self.get_sound(param[0])
                self.default_volume[name] = param[1]

                if sound_obj:
                    sound_obj.set_volume(param[1])
                    if len(param) > 1:
                        sound_obj.set_loop(param[2])
            else:
                self.notify.error('%s.load_sound_data: Unknown param type: %s' % (
                    self.__class__.__name__, param))

    def load_music_data(self, data: dict) -> None:
        """
        Loads music data from configurable variants
        """

        self.default_volume = {}
        for name, param in list(data.items()):
            if type(param) == type(''):
                self.add_music(name, param)
            elif type(param) == tuple:
                self.add_music(name, param[0])

                sound_obj = self.get_sound(param[0])
                self.default_volume[name] = param[1]

                if sound_obj:
                    sound_obj.set_volume(param[1])
                    if len(param) > 1:
                        sound_obj.set_loop(param[2])
            else:
                self.notify.error('%s.load_music_data: Unknown param type: %s' % (
                    self.__class__.__name__, param))

class SoundManager(Singleton, InternalObject):
    """
    Manages all sound sources in the Gemstone framework
    """

    def __init__(self):
        Singleton.__init__(self)
        InternalObject.__init__(self)

        self.__sound_off = False
        self.__sound_sources = {}

    @property
    def sound_off(self):
        """
        Returns the managers currently "off" state as a 
        property
        """

        return self.is_sound_off()

    @sound_off.setter
    def sound_off(self, state):
        """
        Sets the managers current off state
        as a setter
        """

        self.set_sound_off(state)

    def destroy(self):
        """
        Destroys the sound manager singleton instance
        """

        for source_id in self.__sound_sources:
            self.unregister(source_id)

        self.__sound_sources = {}
        InternalObject.destroy(self)

    def __destroy_sound(self, sounds: list, name: str) -> None:
        """
        Destroys a sound name instance from a sounds list.
        """

        if name in sounds:
            del sounds[name]

    def __load_sfx_file(self, path: str) -> object:
        """
        Loads the requested sfx file path from the Panda3d
        Virtual File System
        """

        return runtime.loader.load_sfx(path)

    def __load_music_file(self, path: str) -> object:
        """
        Loads the requested music file path from the Panda3d
        Virtual File System
        """

        return runtime.loader.load_music(path)

    def is_sound_off(self):
        """
        Returns the managers currently "off" state as
        a method.
        """

        return self.__sound_off

    def set_sound_off(self, sound_off):
        """
        Sets the managers current sound off state
        """
    
        self.__sound_off = sound_off

        # If the state is sound off stop all sound sources
        if sound_off:
            for sounds in list(self.__sound_sources.values()):
                for sound_obj, sound_pos in list(sounds.values()):
                    sound_obj.stop()

    def pause_all(self):
        """
        Pauses all sound sources
        """

        for source_id in self.__sound_sources:
            sounds = self.__sound_sources[source_id]

            for name in sounds:
                self.pause(source_id, name)

    def resume_all(self):
        """
        Resumes all sound sources
        """

        for source_id in self.__sound_sources:
            sounds = self.__sound_sources[source_id]

            for name in sounds:
                self.resume(source_id, name)

    def register(self, source_id: object, name: str, filename: str) -> None:
        """
        Registers the sound with its respective source and loads it into 
        memory
        """

        # Verify the sound file exists
        if not path_exists(filename):
            self.notify.error('%s.load: Failed to load sound file %s. Does not exist' % (self.__class__.__name__, filename))
            return

        if source_id not in self.__sound_sources:
            self.__sound_sources[source_id] = {}

        sounds = self.__sound_sources[source_id]
        self.__destroy_sound(sounds, name)
        sounds[name] = (self.__load_sfx_file(filename), -1)

    def unregister(self, source_id: object, name: str = None) -> None:
        """
        """

        # Verify the sound source exists
        if not self.has_sound_source(source_id):
            return

        if name is None:
            sounds = self.get_sound_source(source_id)
            for name in list(sounds.items()):
                sounds[name] = None

            del self.__sound_sources[source_id]
        else:
            self.__destroy_sound(self.__sound_sources[source_id], name)

    def register_music(self, source_id: object, name: str, filename: str) -> None:
        """
        """

        # Verify the sound file exists
        if not path_exists(filename):
            self.notify.error('%s.load: Failed to load sound file %s. Does not exist' % (self.__class__.__name__, filename))
            return

        if source_id not in self.__sound_sources:
            self.__sound_sources[source_id] = {}

        sounds = self.__sound_sources[source_id]
        self.__destroy_sound(sounds, name)
        sounds[name] = (self.__load_music_file(filename), -1)

    def set_music_volume(self, volume: float) -> None:
        """
        Sets the music playback volume
        """

        runtime.base.musicManager.set_volume(volume)

    def set_sound_effects_volume(self, volume: float) -> None:
        """
        Sets the sound effects playback volume
        """

        base = runtime.base
        for i in range(len(base.sfxManagerList)):
            if base.sfxManagerIsValidList[i]:
                base.sfxManagerList[i].setVolume(volume)

    def enable_music(self, state: bool) -> None:
        """
        Sets the music playback enabled state
        """

        runtime.base.enable_music(state)

    def enable_sound_effects(self, state: bool) -> None:
        """
        Sets the sound effects playback enabled state
        """

        runtime.base.enable_sound_effects(state)

    def pause(self, source_id: object, name: str) -> None:
        """
        Pauses the sound from the requested sound source
        """

        if not self.has_sound_source(source_id):
            return

        sounds = self.get_sound_source(source_id)
        if name not in sounds:
            return

        sound_obj, sound_pos = sounds[name]
        if sound_obj.status() == 2:
            sounds[name] = (sound_obj, sound_obj.get_time())
            sound_obj.stop()

    def resume(self, source_id: object, name: str) -> None:
        """
        Resumes the sound from the requested sound source
        """

        if not self.has_sound_source(source_id):
            return

        sounds = self.get_sound_source(source_id)
        if name not in sounds:
            return

        sound_obj, sound_pos = sounds[name]
        if sound_pos != -1:
            sound_obj.set_time(sound_pos)
            sounds[name] = (sound_obj, -1)

    def has_sound_source(self, source_id: object) -> bool:
        """
        Returns true if the given sound source id exists
        """

        return source_id in self.__sound_sources

    def get_sound_source(self, source_id: object) -> object:
        """
        Returns the sound source by id if it exists. Otherwise NoneType
        """

        if not self.has_sound_source(source_id):
            return None

        return self.__sound_sources[source_id]

    def get_sound(self, source_id: object, name: str) -> object:
        """
        Attempts to retrieve a sound name for a given sound source by
        its identifier
        """

        if not self.has_sound_source(source_id):
            return None

        sounds = self.get_sound_source(source_id)
        if name in sounds:
            sound_obj, sound_pos = sounds[name]
            return sound_obj
        
        return None

    def copy_sounds(self, source_id: object, new_id: object) -> None:
        """
        Copys the sounds from one sound source id to another.
        """

        if not self.has_sound_source(source_id):
            return

        sounds = self.get_sound_source(source_id)
        if not self.has_sound_source(new_id):
            self.__sound_sources[new_id] = {}

        self.__sound_sources[new_id].update(sounds)

    def __getattr__(self, name) -> object:
        """
        Custom attribute getter for delegating missing function calls
        to the Panda3D sound object for the respective sound name and source
        """

        return lambda *args: self.__delegate_to_audio_object(name, *args)

    def __delegate_to_audio_object(self, func_name: str, source_id: object, name: str, *args) -> object:
        """
        Delegates the function call from the custom __getattr__ handler to the Panda3D native sound
        object.
        """

        if not self.has_sound_source(source_id):
            self.notify.warning('Failed to delegate audio. Sound source does not exist')
            return

        if self.__sound_off:
            self.notify.debug('Delgation ignored. Sound is currently off')
            return

        if len(self.__sound_sources) == 0:
            self.notify.warning('Failed to delegate audio. No sound sources defined')
            return

        sounds = self.get_sound_source(source_id)
        if name in sounds:
            sound_obj, sound_pos = sounds[name]
            func = getattr(sound_obj, func_name)
            return func(*args)

        raise AttributeError('Count not find attribute "%s" for sound "%s" from source %s.' % (
            func_name, name, source_id))



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.registry import ClassRegistryMixIn
from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework import utilities
from panda3d_gemstone.framework.service import Service, Command

from panda3d_gemstone.logging.utilities import get_notify_category

from panda3d.core import NodePath

__controller_notify = get_notify_category('controller')

class Controller(Configurable, InternalObject):
    """
    Base class for all controllers in the Geomstone framework
    """

    def __init__(self, config_path: str, next_controller: object = None, section = None):
        Configurable.__init__(self, config_path, section or self.__class__.__name__)
        InternalObject.__init__(self)

        self.next_controller = next_controller
        self.moved = False
        self.turned = False
        self.update_nodepath = NodePath('')
        self.initialize()

    @property
    def changed(self) -> bool:
        """
        Value is true if the controller has changed
        """

        return self.has_changed()
    
    def destroy(self) -> None:
        """
        Called on the destruction of the controller object
        """

        self.next_controller = None
        self.update_nodepath.remove_node()
        self.update_nodepath = None

    def set_next_controller(self, next_controller: object) -> None:
        """
        """

        self.next_controller = next_controller

    def get_next_controller(self,) -> object:
        """
        """

        return self.next_controller

    def has_moved(self) -> bool:
        """
        Returns true if the controller has moved
        """

        return self.moved

    def has_turned(self) -> bool:
        """
        Returns true if the controller has turned
        """

        return self.turned

    def has_changed(self) -> bool:
        """
        Returns true if the controller has changed
        """

        return self.moved or self.turned

    def walk_controller_chain(self, ignore_self: bool = False) -> None:
        """
        """

        if ignore_self:
            controller = self.get_next_controller()
        else:
            controller = self

        while controller:
            yield controller
            controller = controller.get_next_controller()

    def foreach_get(self, attr: object) -> list:
        """
        """

        values = []
        for controller in self.walk_controller_chain():
            values.append(getattr(controller, attr, None))

        return values

    def foreach_call(self, method: object, *args, **kwargs) -> list:
        """
        """

        return utilities.foreach_call_method_by_name(self.walk_controller_chain(False), method, *args, **kwargs)

    def _foreach_call_others(self, method: object, *args, **kwargs) -> list:
        """
        """

        return utilities.foreach_call_method_by_name(self.walk_controller_chain(True), method, *args, **kwargs)

    async def update_node(self, nodepath: object, dt: float) -> object:
        """
        """

        return nodepath

    async def forward_node(self, nodepath: object, dt: float) -> object:
        """
        """

        return await self.next_controller.update(nodepath, dt)

    async def update(self, nodepath: object, dt: float) -> object:
        """
        """

        prev_direction = nodepath.get_quat()
        prev_position = nodepath.get_pos()

        new_nodepath = self.update_nodepath
        new_nodepath.set_mat(nodepath.get_mat())
        new_nodepath = await self.update_node(new_nodepath, dt)

        if self.next_controller:
            new_nodepath = await self.forward_node(new_nodepath, dt)

        direction = new_nodepath.get_quat()
        position = new_nodepath.get_pos()

        self.moved = not position.almost_equal(prev_position)
        self.turned = not direction.almost_equal(prev_direction)

        return new_nodepath

class ControllerChainMixIn(ClassRegistryMixIn):
    """
    """

    def call_controller_chain(self, controller_chain: object, function_name: object, *args, **kwargs) -> object:
        """
        """

        if controller_chain:
            return controller_chain.foreach_call(function_name, *argsm **kwargs)

    def setup_controller_chain(self, controller_chain_str: str, config_path: str) -> object:
        """
        """

        first_controller = None

        if controller_chain_str:
            controller_chain = [ c.strip() for c in controller_chain_str.split(' ') ]
            controller_chain.reverse()
            for controller_name in controller_chain:
                cls = self._get_class(controller_name + 'Controller')
                setup = self._get_class_meta(controller_name + 'Controller', 'Controller.setup')
                if cls:
                    first_controller = cls(config_path, next_controller=first_controller)
                    if setup:
                        setup(first_controller, self)
                else:
                    __controller_notify.error("Failed to setup controller '%s' (%s). Class not found" % (controller_name, self.__class__.__name__))
                    return

        return first_controller

    def find_controller(self, controller_chain: object, cls_or_name: object, as_list: bool = False) -> object:
        """
        """

        controller = []
        while controller_chain:
            if isinstance(cls_or_name, str):
                if controller_chain.section == cls_or_name:
                    controller.append(controller_chain)
            elif isinstance(controller_chain, cls_or_name):
                if isinstance(controller_chain, cls_or_name):
                    controller.append(controller_chain)
            controller_chain = controller_chain.get_next_controller()

        if as_list or len(controller) > 1:
            return controller
        elif len(controller) == 1:
            return controller[0]

        return None

class ControllerService(Controller, Service):
    """
    Base class for all controllers who also serve as a Gemstone
    framework Service object for registering commands
    """

    def __init__(self, *args, **kwargs):
        Service.__init__(self)
        Controller.__init__(self, *args, **kwargs)



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import os
import sys
import datetime
import itertools
import datetime
import types
import gc
import functools

from math import floor, ceil
from functools import reduce

from panda3d.core import Filename, Point2, Point3
from panda3d.core import Vec3

from panda3d_gemstone.engine import runtime
from panda3d_gemstone.logging.utilities import get_notify_category
from panda3d_gemstone.framework import exceptions

__utility_notify = get_notify_category('utilities')

_SNAKE_NAME_RE = re.compile('(?<!^)(?=[A-Z])')

def get_snake_case(text: str, splitter='_') -> str:
    """
    Returns the snake case version of the requested string
    """

    return _SNAKE_NAME_RE.sub(splitter, text).lower()

def get_camel_case(text: str, splitter='_') -> str:
    """
    Returns the camel case version of the requested string
    """

    return ''.join(x.capitalize() or splitter for x in text.split(splitter))

def open_web_url(url: str) -> bool:
    """
    Attempts to open the website url. Returning true on sucess
    otherwise False on failure.
    """

    success = False
    if sys.platform == 'darwin':
        os.system('/usr/bin/open %s' % url)
    elif system.platform == 'linux':
        import webbrowser
        webbrowser.open(url)
        success = True
    else:
        try:
            import webbrowser
            webbrowser.open(url)
            success = True
        except:
            os.system('explorer "%s"' % url)

    return success

def get_time_as_string(seconds: bool = False) -> str:
    """
    Returns the current time as a string
    """

    now = datetime.datetime.now()
    output = None
    if seconds:
        output = '%02d:%02d:%02d' % (now.hour, now.minute, now.second)
    else:
        output = '%02d:%02d' % (now.hour, now.minute)

    return output

def is_awaitable_function(func: object) -> bool:
    """
    Returns true if the function is awaitable
    """

    assert func != None
    assert callable(func)

    import inspect
    return inspect.iscoroutinefunction(func)

def run_func_async(func: object, name: str = None) -> None:
    """
    Runs a the requested function under a coroutine
    """

    assert func != None
    assert callable(func)

    if name is None:
        name = '%s-async-task' % func.__name__

    async def async_wrapper(task) -> int:
        """
        Async wrapper for the callback
        """

        await func()
        return task.done

    create_task(async_wrapper, name)

def perform_callback_on_condition(condition: bool, callback: object, *args, **kwargs) -> None:
    """
    Performs the required callback if the condition is True
    """

    assert callback != None
    assert callable(callback)

    # Check if the condition is true
    if not condition:
        return

    if is_awaitable_function(callback):
        run_func_async(callback, '%s-callback-async' % callback.__name__)
    else:
        callback(*args, **kwargs)

def perform_child_class_check(base_class: str, use_cls: object, attr: str) -> None:
    """
    Performs a check to ensure the use class has the required attribute of its base
    class. If the check fails a ImproperClassUsageError is thrown informing
    the user of the issue
    """

    if not hasattr(use_cls, attr):
        raise exceptions.ImproperClassUsageError(base_class, use_cls.__class__.__name__)

def perform_dependency_check(library_name: str, import_path: str = None) -> None:
    """
    Performs a check to ensure the required dependency is installed.
    If the dependency is not found a MissingThirdpartySupportError 
    is thrown to inform the user
    """

    if import_path is None:
        import_path = library_name

    import importlib
    found = False
    try:
        module = importlib.import_module(import_string)
        found = True
    except:
        pass

    if not found:
        raise exceptions.MissingThirdpartySupportError(library_name)

def __create_task_name(obj, task):
    """
    Creates a name for a task based on the obj its being used on
    """

    cls_name = obj.__class__.__name__
    if hasattr(obj, 'get_name'):
        name = obj.get_name()
        return '%s.%s(%s)' % (cls_name, task, name)

    return '%s.%s(<%d>)' % (cls_name, task, id(obj))

def create_task(task_func, task_name = '', priority = 0, task_chain_name: str = None):
    """
    Creates a new task with the task manager
    """

    task_name = __create_task_name(task_func.__self__, task_name or task_func.__name__)
    return runtime.task_mgr.add(task_func, task_name, priority, taskChain=task_chain_name)

def create_delayed_task(task_func, delay, task_name = ''):
    """
    Creates a new delayed task with the task manager
    """

    task_name = __create_task_name(task_func.__self__, task_name or task_func.__name__)
    return runtime.task_mgr.do_method_later(task_func, task_name, priority)

def remove_task(task):
    """
    Removes a task from the task manager
    """

    runtime.task_mgr.remove(task)

def create_thread(thread_name, thread_priority=0):
    """
    Creates a new task manager thread
    """

    runtime.task_mgr.setupTaskChain(
        thread_name, 
        numThreads=1, 
        threadPriority=thread_priority)

def diffs(lst1, lst2):
    """
    """

    return reduce(lambda x, y: x + y, itertools.starmap(lambda e1, e2: int(not e1 == e2), list(zip(lst1, lst2))))


def delegate(self, call) -> None:
    """
    """

    func_name = call.__func__.__name__
    setattr(self, func_name, call)

def null_generator():
    """
    Defines a null yield generator
    """

    if False:
        yield

def open_os_directory(path: str) -> bool:
    """
    Opens a directory path in the operation system's file explorer.
    Returning true on success, otherwise false.
    """

    success = False
    if sys.platform == 'darwin':
        __utility_notify.warning('open_os_directory is not supported on platform: %s' % sys.platform)
    elif sys.platform == 'linux2':
        __utility_notify.warning('open_os_directory is not supported on platform: %s' % sys.platform)
    else:
        os.system('explorer "%s"' % url)
        success = True

    return success

def get_screenshot_directory(absolute: bool = False) -> str:
    """
    Returns the application's screenshot directory
    """

    from panda3d_gemstone.engine import prc
    directory_name = prc.get_prc_string('gs-screenshot-dir', 'screenshots')
    directory_path = '.%s%s' % (os.sep, directory_name)

    if absolute:
        directory_path = os.path.abspath(directory_path)

    return directory_path

def open_screenshot_directory() -> bool:
    """
    Opens the application's screenshot directory in the operation system's
    file browser window. Returning true on success, otherwise false.
    """

    return open_os_directory(get_screenshot_directory(True))

def build_screenshot_filename(basename: str = 'screenshot', directory: str = None, format: str = 'png') -> str:
    """
    Builds the file path for a newly created screenshot
    """

    if directory is None:
        directory = get_screenshot_directory()

    now = datetime.datetime.now()
    filename = now.strftime(basename + '_%y%m%d_%H%M%S')
    path = os.path.join(directory, filename +' .' + format)
    appendix = 0

    while os.path.exists(path):
        appendix += 1
        path = os.path.join(directory, filename + '_' + str(appendix) + '.' + format)

    return path

def save_screenshot(directory: str = None, format: str = 'png', win: object = None):
    """
    Saves a screenshot of the current render output to file
    """

    if not win:
        win = runtime.base.win

    path = build_screenshot_filename(directory=directory, format=format)
    win.save_screenshot(Filename(path))
    __utility_notify.info('Saved Screenshot: %s' % path)

def foreach(sequence: list, callable: object, *args, **kwargs) -> None:
    """
    Iterates through a sequence performing a callback on each element
    """

    for element in sequence:
        callable(element, *args, **kwargs)

def foreach_call_method_by_name(sequence: list, method_name: str, *args, **kw) -> list:
    """
    """

    results = []
    for element in sequence:
        callable = getattr(element, method_name, None)
        if callable:
            results.append(callable(*args, **kw))

    return results

BACKGROUND_THREAD_NAME = 'background-tasks'

def start_background_thread() -> None:
    """
    Starts a background thread chain in the Panda3D
    task manager
    """

    runtime.task_mgr.setupTaskChain(
        BACKGROUND_THREAD_NAME, 
        numThreads = 1,
        threadPriority = 0)

def map_point_to_screen(nodepath: object, point: object) -> object:
    """
    """

    p3 = runtime.cam.get_relative_point(nodepath, point)
    p2 = Point2()

    if not runtime.base.camlens.project(p3, p2):
        return None

    r2d = Point3(p2[0], 0, p2[1])
    a2d = runtime.base.aspect2d.get_relative_point(runtime.base.render2d, r2d)

    return a2d

def snap_to_grid(nodepath: object, grid_size:object) -> tuple:
    """
    """

    x, y, z = node_path[0], node_path[1], node_path[2]
    return (floor(x / grid_size[0]) * grid_size[0], floor(y / grid_size[1]) * grid_size[1], floor(z / gridSize[2]) * grid_size[2])

def get_bounds_of_model(model: object, rotation: float = 0.0) -> tuple:
    """
    """

    h = model.get_h()
    model.set_h(rotation)
    min_corner, max_corner = model.get_tight_bounds()
    model.set_h(h)
    delta = max_corner - min_corner

    return (min_corner, max_corner, Vec3(int(ceil(round(delta.get_x(), 1))), int(ceil(round(delta.get_y(), 1))), int(ceil(round(delta.get_z(), 1)))))

def write_ini_file(filepath: str, input: object, output: object, template: str = '[Configuration]\n\n[Model]\n%(output)s: %(input)s\n') -> None:
    """
    Writes a new ini file to the requested output path location
    """

    __utility_notify.info('write_ini_file: Creating ini file for "%s"... ' % input)
    fh = open(filepath, 'w')
    fh.write(template % {'input': input, 'output': output})
    fh.close()

def has_attributes(object: object, attributes: list) -> list:
    """
    """

    return reduce(bool.__and__, [ hasattr(object, attr) for attr in attributes ])

def get_refcounts() -> list:
    """
    Returns a list of all references in the application
    """

    d = {}
    import sys.modules
    for m in list(sys.modules.values()):
        for sym in dir(m):
            o = getattr(m, sym)
            if type(o) is type:
                d[o] = sys.getrefcount(o)

    pairs = [(refcount, cls) for cls, recount in list(d.items())]
    pairs.sort(reverse=True)

    return pairs

def get_all_references_of_type(t: object) -> list:
    """
    """

    result = []
    for r in gc.get_referrers(t):
        if isinstance(r, t):
            result.append(r)

    return result

def print_refcounts(max_refcounts: int = None) -> None:
    """
    """

    refcount_list = get_refcounts()
    if max_refcounts is not None:
        refcount_list = refcount_list[max_refcounts]
    
    for n, c in refcount_list:
        print('%10d %s (%s) instances: %d' % (
            n, c.__name__, str(c), len(get_all_references_of_type(c))))

def print_unreachable_garbage() -> None:
    """
    """

    garbage_list = gc.garbage
    if len(garbage_list) == 0:
        print('No garbage found')
    else:
        print('%d object in garbage found: ' % len(garbage_list))
        for garbage in garbage_list:
            print(str(garbage))

def to_unicode(string: str) -> str:
    """
    """

    if type(string) == str:
        return string

    try:
        return str(string.decode('utf-8'))
    except UnicodeError:
        return ''

def utf8_capitalize(string: str) -> str:
    """
    """

    return to_unicode(string).capitalize().encode('utf-8')

def utf8_lower(string: str) -> str:
    """
    """

    return to_unicode(string).lower().encode('utf-8')

class _DoMethodAfterNFrames(object):
    """
    """

    def __init__(self, frames_to_wait: int, method: object, args: list):
        self.__frames_to_wait = frames_to_wait
        self.__method = method
        self.__args = args

    def task_function(self, task: object) -> int:
        """
        """

        self.__frames_to_wait -= 1
        
        if self.__frames_to_wait <= 0:
            self.__method(*self.__args)

            return task.done

        return task.cont

def do_method_after_n_frames(frames_to_wait: int, method: object, args: list = [], priority: int = 0) -> None:
    """
    """

    if frames_to_wait > 0:
        create_task(_DoMethodAfterNFrames(frames_to_wait, method, args).task_func, priority=priority)
    else:
        __utility_notify.error('Invalid request. do_method_after_n_frames received a frames wait of 0')

def disallow_frozen(func: object) -> object:
    """
    Wrapper to prevent functions called while in a frozen environment
    """

    from panda3d_gemstone.engine import runtime
    from panda3d_gemstone.framework import exceptions

    def wrapper(*args, **kwargs):
        """
        Wraps the function to perform an environment check
        """

        if runtime.is_frozen():
            raise exceptions.InvalidEnvironmentError(
                '%s is not allowed in a frozen environment' % func.__name__)

        return func(*args, **kwargs)

    return wrapper

def disallow_production(func: object) -> object:
    """
    Wrapper to prevent function called in a production environment
    """

    from panda3d_gemstone.engine import runtime
    from panda3d_gemstone.framework import exceptions

    def wrapper(*args, **kwargs):
        """
        Wraps the function to perform an environment check
        """

        if not runtime.is_developer_build():
            raise exceptions.InvalidEnvironmentError(
                '%s is not allowed in a production environment' % func.__name__)

        return func(*args, **kwargs)

    return wrapper

def disallow_interactive(func: object) -> object:
    """
    Wrapper to prevent function calls in an interactive environment
    """

    from panda3d_gemstone.engine import runtime
    from panda3d_gemstone.framework import exceptions

    def wrapper(*args, **kwargs):
        """
        Wraps the function to perform an environment check
        """

        if runtime.is_interactive:
            raise exceptions.InvalidEnvironmentError(
                '%s is not allowed in an interactive environment' % func.__name__)

        return func(*args, **kwargs)

    return wrapper



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.controller import Controller
from panda3d.core import Vec3, Point3

class FixedController(Controller):
    """
    """

    def __init__(self, config_path: str, next_controller: object = None, section: str = None):
        self.__position = Vec3(0, 0, 50)
        self.__look_at_offset = Vec3(0, 0, 0)
        self.__follow_mode = False
        self.__target = None
        Controller.__init__(self, config_path, next_controller, section or 'Fixed')

    def set_position(self, position: Vec3) -> None:
        """
        """
        self.__position = position

    def get_position(self) -> Vec3:
        """
        """

        return self.__position

    def set_look_at_offset(self, look_at_offset: Vec3) -> None:
        """
        """

        self.__look_at_offset = look_at_offset

    def get_look_at_offset(self) -> Vec3:
        """
        """

        return self.__look_at_offset

    def set_follow_mode(self, follow: bool) -> None:
        """
        """

        self.__follow_mode = follow

    def get_follow_mode(self) -> bool:
        """
        """

        return self.__follow_mode

    def set_target(self, target: object) -> None:
        """
        """

        self.__target = target

    def get_target(self) -> object:
        """
        """

        return self.__target

    async def update(self, nodepath: object, dt: float) -> object:
        """
        """

        nodepath.set_pos(self.__position)
        if self.__target and self.__follow_mode:
            if hasattr(self.__target, '__getitem__'):
                look_at = Point3(self.__target[0], self.__target[1], self.__target[2])
            else:
                look_at = Point3(self.__target.getX(), self.__target.getY(), self.__target.getZ())
        else:
            look_at = Point3(0, 0, 0)


        nodepath.look_at(look_at + self.__look_at_offset, Vec3(0, 0, 1))
        return nodepath



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import traceback

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.runnable import Runnable

from panda3d.core import HTTPClient, HTTPChannel, DocumentSpec
from panda3d.core import Ramfile, UniqueIdAllocator, ConfigVariableInt

class HTTPManager(Singleton, Runnable, InternalObject):
    """
    Singleton object for handling GET/POST requests with Panda3D internal
    HTTPClient object
    """

    def __init__(self):
        Singleton.__init__(self)
        Runnable.__init__(self)
        InternalObject.__init__(self)

        self._http_client = HTTPClient()

        max_http_requests = ConfigVariableInt('http-max-requests', 900).value
        self._request_allocator = UniqueIdAllocator(0, max_http_requests)
        self._poll_task = None
        self._requests = {}

        self.activate()

    async def tick(self, dt: float) -> None:
        """
        Called once per frame to update any pending HTTP request
        objects
        """

        for request_id in list(self._requests):

            # Check that this id is still valid
            if request_id not in self._requests:
                continue

            request = self._requests[request_id]
            await request.tick()

    def destroy(self) -> None:
        """
        Destroys the http manager object
        """

        Singleton.destroy(self)
        InternalObject.destroy(self)

        for request_id in list(self._requests):
            self.remove_request(request_id)

        self.deactivate()

    def remove_request(self, request_id: object) -> None:
        """
        Removes the request id form the PandaHTTP request list
        """
        
        if request_id not in self._requests:
            return

        self._request_allocator.free(request_id)
        del self._requests[request_id]

    def get_request_status(self, request_id: object) -> bool:
        """
        Returns the requests current status
        """

        return not request_id in self._requests

    def get_request(self, request_id: object) -> object:
        """
        Returns the requested request if its present
        """

        return self._requests.get(request_id, None)

    def perform_get_request(self, url: str, headers: dict = {}, content_type: str = None, callback: object = None) -> object:
        """
        Performs an HTTP restful GET call and returns the request's unique itentifier
        """

        self.notify.debug('Sending GET request: %s' % url)

        request_channel = self._http_client.make_channel(True)

        if content_type != None:
            request_channel.set_content_type(content_type)

        for header_key in headers:
            header_value = headers[header_key]
            request_channel.send_extra_header(header_key, header_value)

        request_channel.begin_get_document(DocumentSpec(url))

        ram_file = Ramfile()
        request_channel.download_to_ram(ram_file, False)

        request_id = self._request_allocator.allocate()
        http_request = HTTPRequest(self, request_id, request_channel, ram_file, callback)
        self._requests[request_id] = http_request

        return request_id

    def perform_json_get_request(self, url: str, headers: dict = {}, callback: object = None) -> object:
        """
        """

        def json_wrapper(data: str) -> None:
            """
            Wraps the callback to automatically perform json.load
            on the resulting data
            """

            try:
                data = json.loads(data)
            except:
                self.notify.warning('Received invalid JSON results: %s' % data)
                
            callback(data)

        return self.perform_get_request(
            url=url, 
            content_type='application/json',
            headers=headers,
            callback=json_wrapper)

    def perform_post_request(self, url: str, headers: dict = {}, content_type: str = None, post_body: dict = {}, callback: object = None) -> object:
        """
        """

        self.notify.debug('Sending POST request: %s' % url)

        request_channel = self._http_client.make_channel(True)

        if content_type != None:
            request_channel.set_content_type(content_type)

        for header_key in headers:
            header_value = headers[header_key]
            request_channel.send_extra_header(header_key, header_value)

        post_body = json.dumps(post_body)
        request_channel.begin_post_form(DocumentSpec(url), post_body)

        ram_file = Ramfile()
        request_channel.download_to_ram(ram_file, False)

        request_id = self._request_allocator.allocate()
        http_request = HTTPRequest(self, request_id, request_channel, ram_file, callback)
        self._requests[request_id] = http_request

        return request_id

    def perform_json_post_request(self, url: str, headers: dict = {}, post_body: dict = {}, callback: object = None) -> object:
        """
        """

        def json_wrapper(data: str) -> None:
            """
            Wraps the callback to automatically perform json.load
            on the resulting data
            """

            try:
                data = json.loads(data)
            except:
                self.notify.warning('Received invalid JSON results: %s' % data)

            callback(data)

        return self.perform_post_request(
            url=url, 
            content_type='application/json',
            headers=headers,
            post_body=post_body, 
            callback=json_wrapper)

    def get_notify_name(self) -> str:
        """
        Returns this object's notifier name
        """

        return 'http-manager'

class HTTPRequest(InternalObject):
    """
    Represents an inprogres HTTP request
    """

    def __init__(self, rest: HTTPManager, request_id: object, channel: object, ram_file: object, callback: object = None):
        super().__init__()
        self._rest = rest
        self._request_id = request_id
        self._channel = channel
        self._callback = callback
        self._ram_file = ram_file
    
    @property
    def request_id(self) -> object:
        return self._request_id

    @property
    def channel(self) -> object:
        return self._channel

    @property
    def ram_file(self) -> object:
        return self._ram_file

    async def tick(self) -> None:
        """
        Performs the run operations and finishing callbacks
        for the request's channel instance
        """

        if self._channel == None:
            return

        done = not self._channel.run()
        if done:
            self.notify.debug('Completed request: %s' % self._request_id)
            
            if self._callback != None:
                try:
                    self._callback(self._ram_file.get_data())
                except:
                    self.notify.warning('Exception occured processing callback')
                    self.notify.warning(traceback.format_exc())

            self._rest.remove_request(self._request_id)

    def get_notify_name(self) -> str:
        """
        Returns this object's notifier name
        """

        return 'http-request'


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys

from panda3d.core import MultiplexStream, Filename, load_prc_file_data
from panda3d.core import Notify, ConfigVariableString

from panda3d_gemstone.engine import runtime

def get_notify_categories() -> object:
    """
    Retrieves all Panda3D notifier categories
    """

    from direct.directnotify.DirectNotifyGlobal import directNotify
    return directNotify.getCategories()

def get_notify_category(name: str, create: bool = True) -> object:
    """
    Returns the requested Panda3D notifier category. Creating a new
    one if create is set to True
    """

    assert name != None
    assert name != ''

    from direct.directnotify.DirectNotifyGlobal import directNotify

    category = None
    if create:
        category = directNotify.newCategory(name)
    else:
        category = directNotify.getCategory(name)
    return category

def log(message: str, name: str = 'global', type: str = 'info') -> None:
    """
    Writes a message to the requested logger name
    """

    category = get_notify_category(name)
    assert hasattr(category, type)
    getattr(category, type)(message)

def log_error(message: str, name: str = 'global') -> None:
    """
    Writes an error message to the requested logger name
    """

    log(name, message, 'error')

def log_warn(message: str, name: str = 'global') -> None:
    """
    Writes an warn message to the requested logger name
    """

    log(name, message, 'warn')

def log_info(message: str, name: str = 'global') -> None:
    """
    Writes an info message to the requested logger name
    """

    log(name, message, 'info')

def log_debug(message: str, name: str = 'global') -> None:
    """
    Writes an debug message to the requested logger name
    """

    log(name, message, 'debug')

def condition_error(logger: object, condition: bool, message: str) -> None:
    """
    Writes a error message to the logging object if the provided
    condition is true
    """

    condition_log(logger, condition, message, 'error')

def condition_warn(logger: object, condition: bool, message: str) -> None:
    """
    Writes a warning message to the logging object if the provided
    condition is true
    """

    condition_log(logger, condition, message, 'warning')

def condition_info(logger: object, condition: bool, message: str) -> None:
    """
    Writes a info message to the logging object if the provided
    condition is true
    """

    condition_log(logger, condition, message, 'info')

def condition_debug(logger: object, condition: bool, message: str) -> None:
    """
    Writes a debug message to the logging object if the provided
    condition is true
    """

    condition_log(logger, condition, message, 'debug')

def condition_log(logger: object, condition: bool, message: str, type: str = 'info') -> None:
    """
    Writes a message to the logging object if the provided
    condition is true using the supplied type attribute function name
    """

    assert hasattr(logger, type)
    if condition:
        getattr(logger, type)(message)

def get_log_directory() -> str:
    """
    Returns this applications log directory
    based on the PRC configuration
    """

    default = '.%slogs' % os.sep
    return ConfigVariableString('gs-log-directory', default).value

def get_log_filename() -> str:
    """
    Returns the filename used for this applications
    log file
    """

    basename = runtime.executable_name.lower()
    log_ext = ConfigVariableString('gs-log-ext', 'txt').value
    return '%s.%s' % (basename, log_ext)



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import copy
import weakref
import codecs

from panda3d_gemstone.framework.pcast import cast
from panda3d_gemstone.logging.utilities import get_notify_category
from panda3d_gemstone.engine import prc

WORD = '([A-Za-z_][a-zA-Z_0-9]*)'
OPTIONAL_WORD = WORD + '?'
LINE_RE = re.compile('(\\s*)(\\*?)\\s*' + WORD + '\\s*' + OPTIONAL_WORD + '\\s*\\:\\s*(.*)')
__notify = get_notify_category('indented-file')
__detailed_syntax_error = prc.get_prc_bool('gs-if-detailed-syntax', False)

class ParserNode(object):
    """
    Represents a node in the indented file parser
    """

    def __init__(self, pn: object = None):
        if pn:
            self.type = pn.type
            self.free = pn.free
            self.name = pn.name
            self.args = copy.copy(pn.args)
            self.parent = pn.parent
        else:
            self.type = ''
            self.free = ''
            self.name = ''
            self.args = ''
            self.parent = None

    def __str__(self):
        """
        Custom string handler for printing out information regarding
        the ParserNode object
        """

        if self.parent:
            parent_name = self.parent.name
            if not parent_name:
                parent_name = 'ID: ' + str(id(self.parent))
        else:
            parent_name = 'None'

        return 'self: %s\ntype: %s\nfree: %s\nname: %s\nargs: %s\nparent: %s' % (
            id(self), self.type, self.free, self.name, self.args, parent_name)

def __strip_comment(line: str, comment_head: object) -> str:
    """
    """

    comment_found = line.find(comment_head)
    if comment_found != -1:
        line = line[:comment_found]
    
    return line

def parse(path: str, cast_args: bool = False) -> list:
    """
    Parses a Indented file from the application's file
    system
    """

    line_nr = 0
    parents = []
    parents_indent = [-1]
    nodes = []
    in_file = None

    if path:
        in_file = codecs.open(path, 'r', 'utf-8')

    if not in_file:
        __notify.error('Failed to open indented file: %s' % path)
        return nodes

    for line in in_file:
        line_nr += 1

        line = __strip_comment(line, '#')
        line = __strip_comment(line, ';')

        if line.strip():
            m = LINE_RE.match(line)

            if not m:
                if __detailed_syntax_error:
                    __notify.warning('Syntax error in file "%s"; Line: %s (%d)' % (path, line, line_nr))
                else:
                    __notify.warning('Syntax error in file "%s"; Line: %d' % (path, line_nr))
                continue

            pn = ParserNode()
            nodes.append(pn)

            indent, pn.free, pn.type, pn.name, pn.args = m.groups()
            if cast_args:
                pn.args = cast(pn.args)
            
            indent = len(indent)
            while indent <= parents_indent[-1]:
                parents.pop()
                parents_indent.pop()

            if len(parents) > 0:
                pn.parent = parents[-1]
            
            parents.append(pn)
            parents_indent.append(indent)
    
    return nodes

def build_child_list(nodes: list) -> list:
    """
    """

    for n in nodes:
        n.children = []
        if n.parent:
            n.parent.children.append(weakref.ref(n))

    return nodes

def copy_node_list(nodes: list) -> list:
    """
    """

    ret_nodes = []
    for idx, template_node in enumerate(nodes):
        node = ParserNode(template_node)
        if node.parent and node.parent in nodes:
            parent_index = nodes.indeX(node.parent)
            node.parent = ret_nodes[parent_index]

        ret_nodes.append(node)

    ret_nodes = build_child_list(ret_nodes)
    return ret_nodes

def get_indent(node: object) -> int:
    """
    """

    if node.parent:
        return get_indent(node.parent) + 1
    return 0

def print_node_list(nodes: list) -> None:
    """
    """

    for node in nodes:
        indent = get_indent(node)
        print(' ' * indent * 4 + node.type + ':', node.args)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.logging import utilities as __logging
import importlib as __importlib

__bootstrap_notify = __logging.get_notify_category('bootstrap')

def create_class_entry(object_path: str, meta: dict = {}) -> tuple:
    """
    Creates a class entry for use with the bootstrap function
    """

    parts = object_path.split('.')
    class_name = parts[-1]

    return (class_name, object_path, meta)

def create_singleton_entry(object_path: str, meta: dict = {}) -> tuple:
    """
    Creates a singleton entry for use with the bootstrap function
    """

    parts = object_path.split('.')
    class_name = parts[-1]
    path = '.'.join(parts[:-1])

    return (path, class_name, meta)

def get_class_registry() -> object:
    """
    Retrieves the environments class registry
    singleton object
    """

    from panda3d_gemstone.framework.registry import ClassRegistry
    return ClassRegistry.instantiate_singleton()

def __import_module(module_name: str) -> object:
    """
    Imports the requested module by its module
    name import path
    """

    assert module_name != ''
    assert module_name != None

    components = module_name.split('.')
    module_path = '.'.join(components[:-1])
    module = __importlib.import_module(module_path)

    return module

def batch_instantiate_singletons(singleton_list: list) -> None:
    """
    Batch instantiates singletons from a list
    """

    for singleton in singleton_list:
        module_name, class_name, args = singleton
        module_path = '%s.%s' % (module_name, class_name)
        singleton_module = __import_module(module_path)
        if singleton_module is None:
            __bootstrap_notify.warning('Failed to setup singleton: %s. Invalid import' % class_name)
            continue

        singleton_cls = getattr(singleton_module, class_name)
        singleton_cls.instantiate_singleton(*args)

def __verify_thirdparty(modules: object) -> None:
    """
    Verifies the required modules are found in the environment
    prior to importing the rest of the gemstone module
    """

    from panda3d_gemstone.framework import exceptions
    exceptions.verify_thirdparty_modules(modules)

def bootstrap_module(class_list: list = [], meta_list: list = [], singleton_list: list = [], thirdparty: object = []) -> None:
    """
    Performs initial boostrap operations on a module
    """

    class_registry = get_class_registry()
    batch_instantiate_singletons(singleton_list)
    class_registry.batch_register_classes(class_list, meta_list)
    __verify_thirdparty(thirdparty)



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.engine import prc

def get_gemstone_modules() -> list:
    """
    Returns all modules used by all gemstone applications
    """

    return [
        'panda3d_gemstone.application',
        'panda3d_gemstone.controllers',
        'panda3d_gemstone.engine',
        'panda3d_gemstone.framework',
        'panda3d_gemstone.game',
        'panda3d_gemstone.gui',
        'panda3d_gemstone.io',
        'panda3d_gemstone.network',
        'panda3d_gemstone']

def __file_exists(path: str) -> bool:
    """
    Returns true if the file exists
    """

    return os.path.exists(path)

class CompilerError(RuntimeError):
    """
    Base error for all compiler errors
    """

class InvalidCompileError(CompilerError):
    """
    Thrown when an invalid compile operation occures
    """

class BuildTarget(object):
    """
    Represents a build target for the Panda3D compiler
    """

    def __init__(self, key: str, executable: str, includes: list = [], modules: list = [], icons: list = [], plugins: list = []):
        self.__key = key
        self.__executable = executable
        self.__includes = includes
        self.__modules = modules
        self.__icons = icons
        self.__plugins = plugins

    @property
    def key(self) -> str:
        """
        Returns this build target's key
        """

        return self.__key

    @property
    def executable(self) -> str:
        """
        Returns this build target's executable file
        """

        return self.__executable

    @property
    def includes(self) -> list:
        """
        Returns this build target's include list
        """

        return self.__includes

    @property
    def modules(self) -> list:
        """
        Returns this build target's module list
        """

        return self.__modules

    @property
    def icons(self) -> list:
        """
        Returns this build target's icon list
        """

        return self.__icons

    @property
    def plugins(self) -> list:
        """
        Returns this build target's required plugins
        """

        return self.__plugins

    def is_valid(self) -> bool:
        """
        Returns true if this build target is valid
        """

        valid = True

        # Verify the executable file exists
        if not __file_exists(self.__executable):
            valid = False

        return valid

class GemstoneCompiler(InternalObject):
    """
    Base tool for compiling Gemstone applications using
    the Panda3D game engine build system
    """

    def __init__(self, application: str, prc: list = [], debug: bool = False):
        InternalObject.__init__(self)
        self.__application = application
        self.__prc = prc
        self.__debug = debug
        self.__targets = []

    @property
    def targets(self) -> list:
        """
        Returns this compiler objects build targets
        """

        return self.__targets

    def __load_prc_files(self) -> None:
        """
        Loads the compilers prc files
        """

        for prc_file in self.__prc:
            prc.load_prc_file(prc_file)

    def setup(self) -> None:
        """
        Performs setup operations on the compiler
        object instance
        """

        self.__load_prc_files()

    def destroy(self) -> None:
        """
        Performs the compiler objects destruction 
        operations
        """

        InternalObject.destroy(self)

    def add_build_targets(self, targets: list) -> None:
        """
        Adds a list of build targets to the compiler
        """

        for target in targets:
            self.add_build_target(target)

    def get_build_target_by_key(self, key: str) -> BuildTarget:
        """
        Registers a list of build targets with the compiler
        """

        target = None
        for target in self.__targets:
            if target.key == key:
                found = target
                break
        
        return target

    def add_build_target(self, target: BuildTarget) -> None:
        """
        Registers a new build target with the compiler
        """

        # Verify a target does not already exist
        if self.get_build_target_by_key(target.key) != None:
            raise InvalidCompileError('Failed to add target: %s. Target already exists with that key' % target.key)

        self.__targets.append(target)

    def __get_include_patterns(self) -> dict:
        """
        Gets the include patterns required by this compiler's
        build targets
        """

        includes = {}
        include_count = 0
        for target in self.__targets:
            includes[target.key] = target.includes
            include_count += len(target.includes)

        if include_count == 0:
            self.notify.warning('No includes identified. You may experience unexpected behaviour')

        return includes

    def __get_include_modules(self) -> dict:
        """
        Gets the include modules required by this compiler's
        build targets
        """

        modules = {}
        module_count = 0
        for target in self.__targets:
            modules[target.key] = target.modules
            module_count += len(target.modules)

        if module_count == 0:
            self.notify.warning('No modules identified. You may experience unexpected behaviour')

        return modules

    def __get_gui_apps(self) -> dict:
        """
        Gets the gui apps required by this compiler's 
        build targets
        """

        apps = {}
        for target in self.__targets:
            apps[target.key] = target.executable

        if len(apps) == 0:
            raise InvalidCompileError('Failed to compile %s. No GUI apps defined' % self.__class__.__name__)

        return apps

    def __get_icons(self) -> dict:
        """
        Gets the icons required by this compiler's 
        build targets
        """

        icons = {}
        icon_count = 0
        for target in self.__targets:
            icons[target.key] = target.icons
            icon_count += len(target.icons)

        if icon_count == 0:
            self.notify.warning('No icons identified. You may experience unexpected behaviour')

        return icons 

    def __get_plugins(self) -> list:
        """
        Gets the plugins required by the compiler's 
        build targets
        """

        identified = []
        for target in self.__targets:
            for plugin in target.plugins:
                if plugin not in identified:
                    identified.append(plugin)

        if len(identified) == 0:
            self.notify.warning('No plugins identified. You may experience unexpected behaviour.')

        return identified

    def __get_log_filename(self) -> str:
        """
        Returns the log filename used by the compiler
        """

        return prc.get_prc_string('gs-compiler-log', './%s-output.txt' % self.__class__.__name__)

    def __get_build_options(self) -> dict:
        """
        Returns the setup options used by the application
        to compile the Gemstone application
        """

        options = {
            'include_patterns': self.__get_include_patterns(),
            'include_modules': self.__get_include_modules(),
            'gui_apps': self.__get_gui_apps(),
            'icons': self.__get_icons(),
            'log_filename': self.__get_log_filename(),
            'log_append': prc.get_prc_bool('gs-compiler-log-append', False),
            'plugins': self.__get_plugins()
        }

        if self.__debug:
            print('\n==== Compiler Options ====\n')
            self.__dump_dict_to_console(options)
            print('\n')

        return options

    def __dump_dict_to_console(self, data: dict) -> None:
        """
        Dumps a dictionary to console for debugging
        """

        import json
        print(json.dumps(data, indent=4))

    def __get_compiler_setup(self) -> object:
        """
        Returns the setup tools setup object
        for use in the Gemstone application
        compiling process
        """

        from setuptools import setup
        return setup

    def __get_application_version(self) -> str:
        """
        Returns this application's version
        """

        from panda3d_gemstone.application.version import ApplicationVersion
        version_inst = ApplicationVersion.instantiate_singleton()

        return '%s.%s.%s' % version_inst.get()

    def compile(self) -> object:
        """
        Performs the application compiling process.
        Intended to be called from the setup file
        in the Gemstone applications root directory
        """

        if len(self.__targets) == 0:
            raise InvalidCompileError('Failed to compiler %s. No build targets found' % (
                self.__application))

        setup_cls = self.__get_compiler_setup()
        compiler_setup = setup_cls(
            name=self.__application,
            version=self.__get_application_version(),
            options={'build_apps': self.__get_build_options()})

        return compiler_setup




"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

class DataContainer(object):
    """
    Base class for all gemstone data container objects
    """

    def __init__(self, initial: dict = {}):
        for key, value in initial.items():
            self[key] = value

    def get(self, *args, **kwargs) -> object:
        return self.__dict__.get(*args, **kwargs)

    def as_dict(self):
        return self.__dict__

    def __getattr__(self, key: str):
        return self[key]

    def __setattr(self, key: str, item: object):
        self[key] = item

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, key):
        return key in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def __cmp__(self, dict_):
        return self.__cmp__(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __unicode__(self):
        return unicode(repr(self.__dict__))

    def __str__(self):
        return '%s (%s items)' % (self.__class__.__name__, len(self))


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d.core import NodePath, Point3

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.registry import ClassRegistryMixIn
from panda3d_gemstone.framework.cast import cast
from panda3d_gemstone.io.file_system import fix_path

class EntityProxy(ClassRegistryMixIn, InternalObject):
    """
    """

    @staticmethod
    def from_string(s, lineNr = 0) -> object:
        """
        """

        entityProxy = EntityProxy('', '', '', '', '', '', Point3(0, 0, 0), 0, {})
        if entityProxy.deseralize(s):
            return entityProxy
        
        return None

    def __init__(self, layer_id, cls_name, config_path, name, model_cls_name, model_config_path, pos, heading, data):
        InternalObject.__init__(self)
        self.layer_id = layer_id
        self.cls_name = cls_name
        self.config_path = config_path
        self.name = name
        self.model_cls_name = model_cls_name
        self.model_config_path = model_config_path
        self.pos = pos
        self.heading = heading

        if not isinstance(data, dict):
            data = {}
            self.notify.error("Dictionary expected - data of entity '%s' is broken" % name)
        
        self.data = data

    def seralize(self) -> None:
        """
        """

        return "%s <%s> '%s' '%s' %s '%s' %f %f %f %f %s" % (
            self.cls_name,
            self.layer_id,
            self.name,
            fix_path(self.config_path),
            self.model_cls_name,
            fix_path(self.model_config_path),
            self.pos.getX(),
            self.pos.getY(),
            self.pos.getZ(),
            self.heading,
            {})

    def deseralize(self, s) -> bool:
        """
        """

        elements = s.split(' ')
        if len(elements) < 10:
            return False

        self.cls_name = elements[0]
        self.layer_id = elements[1].strip(' <>')
        self.name = elements[2].strip(" '")
        self.config_path = elements[3].strip(" '")
        self.model_cls_name = elements[4].strip(' ')
        self.model_config_path = elements[5].strip(" '")
        self.pos = Point3(float(elements[6]), float(elements[7]), float(elements[8]))
        self.heading = float(elements[9])
        self.data = cast(' '.join(elements[10:]))

        return True

    def instantiate(self, parent, lineNr = 0) -> object:
        """
        """

        cls = self._get_class(self.cls_name)

        if self.model_cls_name != 'None':
            model_cls = self._get_class(self.model_cls_name)
        else:
            model_cls = None
        
        if not cls:
            self.notify.error("Failed to instantiate entity. Unknown entity class '%s' at line #%d", self.cls_name, lineNr)
        elif not model_cls and self.model_cls_name != 'None':
            self.notify.error("Failed to instantiate entity. Unknown entity model class '%s' at line #%d", self.model_cls_name, lineNr)
        else:
            entity = cls(self.config_path)
            entity.set_name(self.name)
            entity.reparent_to(parent, layer_id=self.layer_id)
            entity.set_pos(self.pos)
            entity.set_h(self.heading)
            #entity.set_configuration(self.data)
            if model_cls:
                entity.set_model(model_cls(self.model_config_path))
        
        if not entity:
            entity = NodePath('')

        return entity


class Entity(NodePath, InternalObject, Configurable):
    """
    Base class for all entities in the gemstone framework
    """

    def __init__(self, config_path):
        self._layer = None
        self._model = None
        self._user_data = {}
        InternalObject.__init__(self)
        Configurable.__init__(self, config_path)
        NodePath.__init__(self, '')
        self.initialize()

    @property
    def layer(self) -> object:
        """
        Returns this entity's layer as a property
        """

        return self.get_layer()

    @property
    def world(self) -> object:
        """
        Returns this entity's world as a property
        """

        return self.get_world()

    @property
    def model(self) -> object:
        """
        Returns this entity's model as a property
        """

        return self.get_model()

    @property
    def user_data(self) -> dict:
        """
        Returns this entity's user data as a property
        """

        return self.get_user_data()

    @property
    def rendered(self) -> bool:
        """
        Value is true when the Entity is currently being rendered
        """

        return self.is_rendered()

    def destroy(self) -> None:
        """
        Performs destruction operations on the entity
        """

        PandaBaseObject.destroy(self)
        self.clear_model()
        self.remove_node()

    def is_rendered(self) -> bool:
        """
        Retrusn true if the Entity is currently
        being rendered
        """

        return not self.get_parent().is_empty()

    def set_name(self, name: str) -> None:
        """
        Sets the entity's scene graph name
        """

        # Sanitize the incoming name if it contains
        # whitespaces
        if ' ' in name:
            self.notify.warning('Spaces removed from entity name: "%s' % name)
            name = name.replace(' ', '')

        NodePath.set_name(self, name)

    def get_layer(self) -> object:
        """
        Return's this entity's layer
        """

        return self._layer

    def get_world(self) -> object:
        """
        Returns this entity's world instance
        """

        return self._layer.get_world()

    def reparent_to(self, root: object, sort: int = 0, layer_id: str = '') -> None:
        """
        Reparents this entity to a new node parent
        """

        self.detach_node()

        if layer_id is not None and hasattr(root, 'get_layer'):
            self._layer = root.get_layer(layer_id, create_if_needed=True)
        
        if self._layer:
            self._layer.add_entity(self)

        NodePath.reparent_to(self, root, sort)

    def clear_model(self) -> None:
        """
        Clears this entity's model instance
        """

        if self._model is not None:
            self._model.destroy()
            self._model = None

    def get_model(self) -> object:
        """
        Returns this entity's model instance
        """

        return self._model

    def set_model(self, model_template: object) -> None:
        """
        Sets this entity's model from a model template
        """

        self.clear_model()
        if hasattr(model_template, 'copyTo'):
            self._model = model_template.copy_to(self)

    def attach_model(self, model: object) -> None:
        """
        Ataches a model to the entity
        """

        self.clear_model()
        self._model = model
        model.reparent_to(self)

    def copy_model_to(self, root: object) -> None:
        """
        Copys the entity model to the new root
        """

        assert root != None
        return self._model.copy_to(root)

    def copy_to(self, parent: object) -> object:
        """
        """

        copy = self.__class__(self.path)
        copy.reparent_to(parent, layer_id=self._layer)
        copy.set_pos(self.get_pos())
        copy.set_h(self.get_h())

        if self.get_model() and not copy.get_model():
            copy.set_model(self.get_model())

        if self.get_name() and not copy.get_name():
            copy.set_name(self.get_name())

        return copy

    def __remove_from_layer(self) -> None:
        """
        Removes the entity instance from its current layer
        """

        if self._layer:
            self._layer.remove_entity(self)
            self._layer = None

    def remove_node(self) -> None:
        """
        Removes the entity node from the scene graph
        """

        self.__remove_from_layer()
        NodePath.remove_node(self)

    def detach_node(self) -> None:
        """
        Detaches the entity node from the scene graph
        """

        self.__remove_from_layer()
        NodePath.detach_node(self)

    def set_user_data(self, user_data: dict) -> None:
        """
        Sets the entity's user data
        """

        if isinstance(user_data, dict):
            self._user_data = user_data
        else:
            self.notify.error('Setting user data failed. Data is not a valid dictionary')

    def get_user_data(self) -> dict:
        """
        Returns this entity's user data
        """

        return self._user_data

    def update_user_data(self, user_data: dict) -> None:
        """
        Updates the entity's user data with another 
        dictionary
        """

        self._user_data.update(user_data)

    def set_user_data(self, key: str, value: object) -> None:
        """
        Sets the entity's user data key with the
        requested value
        """

        self._user_data[key] = value

    def get_user_data(self, key: str, default: object = None) -> object:
        """
        Returns this entity's user data key
        if present. otherwise returns the value
        of default
        """

        return self._user_data.get(key, default)

    def set_configuration(self, data: dict) -> None:
        """
        """

        if data:
            self.configuration = data
            Configurable.initialize(self)

    def seralize_configuration(self) -> dict:
        """
        """

        if self._user_data:
            return {'userData': self._user_data}
        else:
            return {}

    def attempt_reload(self) -> None:
        """
        Attempts to reload the entity object
        """

    def __str__(self) -> str:
        """
        """

        layer_id = ''
        if hasattr(self, '__layer'):
            layer_id = self.__layer.get_id()
        
        path = ''
        if hasattr(self, 'path'):
            path = self.path
        
        model_cls_name = 'None'
        model_config_path = ''
        if hasattr(self, '__model'):
            model_cls_name = self.__model.__class__.__name__
            model_config_path = self.__model.path
        
        return EntityProxy(
            layer_id, 
            self.__class__.__name__, 
            path, 
            self.get_name(), 
            model_cls_name, 
            model_config_path, 
            self.get_pos(), 
            self.get_h(), 
            self.seralize_configuration()).seralize()


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.engine import runtime, shader
from panda3d_gemstone.engine.model import StaticModel

from panda3d_gemstone.framework.runnable import Runnable


class Skybox(StaticModel, Runnable):
    """
    Base class for all skybox models
    """

    def __init__(self, config_path: str):
        StaticModel.__init__(self, config_path)
        Runnable.__init__(self, 15)
        
        self.set_bin('background', 0)
        self.set_depth_test(False)
        self.set_compass()
        self.activate()

    def set_lights_enabled(self, state: bool) -> None:
        """
        Sets the lights enable state of the SkyBox
        """

        if not state:
            self.set_light_off()

    def activate(self) -> None:
        """
        Custom activate handle to control the clear color active 
        state with the Panda3D game window
        """

        active = Runnable.activate(self)
        if active and runtime.base.win:
            runtime.base.win.set_clear_color_active(False)

    def deactivate(self) -> None:
        """
        Custom deactivate handle to control the clear color active 
        state with the Panda3D game window
        """

        active = Runnable.deactivate(self)
        if active and runtime.base.win:
            runtime.base.win.set_clear_color_active(True)

    async def tick(self, dt: float) -> int:
        """
        Custom tick handler. Updates the position of the skybox to match
        the position of the users camera
        """

        camera = runtime.camera
        if runtime.base.is_oobe():
            camera = runtime.base.oobeCamera
        self.set_pos(camera.get_pos(self.get_parent()))


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

def verify_thirdparty_modules(modules: object) -> None:
    """
    Verifies the required thirdparty modules are available
    in the environment. If the modules are not found
    a MissingThirdpartySupportError is thrown
    """

    if isinstance(modules, list):
        module_list = modules
    else:
        module_list = [modules]

    from panda3d_gemstone.engine import runtime
    for module in module_list:
        found = runtime.has_thirdparty(module)
        if not found:
            raise MissingThirdpartySupportError(module)

def verify_attribute(obj: object, attrib_name: str) -> None:
    """
    Raises an AttributeError if the object does not have the 
    required attribute name present
    """

    object_name = obj.__class__.__name__
    found = hasattr(obj, attrib_name)
    if not found:
        raise AttributeError('%s does not have the required attribute: %s' % (
            object_name, attrib_name))

def raise_not_implemented(obj: object, method_name: str) -> None:
    """
    Raises a NotImplementedError relating to inform the user
    of the missing method in a child class
    """

    if isinstance(obj, str):
        obj_name = obj
    else:
        obj_name = obj.__class__.__name__

    message = '%s does not implement %s.' % (obj_name, method_name)
    raise NotImplementedError(message)

class ImproperClassUsageError(RuntimeError):
    """
    Thrown when a parent class is improperly used
    with an unsupported child class
    """

    def __init__(self, base_class: str, use_class: str):
        super().__init__('Improper use of %s! %s is not a valid child class.' % (base_class, use_class))

class MissingThirdpartySupportError(RuntimeError):
    """
    Thrown when a thirdparty dependency is missing
    from the environment
    """

    def __init__(self, library_name):
        super().__init__('Missing required thirdparty dependency: %s' % library_name)

class NotSupportedError(RuntimeError):
    """
    Thrown when a method is used that is not supported by
    a variant of the base class
    """

    def __init__(self, cls: object, method_name: str):
        super().__init__('%s does not support %s.' % (cls.__name__, method_name))

class InvalidEnvironmentError(RuntimeError):
    """
    Thrown when an invalid environment is detected
    """


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys

import logging
from logging import StreamHandler

from direct.directnotify.Notifier import Notifier
from panda3d.core import MultiplexStream, Filename, load_prc_file_data
from panda3d.core import Notify, ConfigVariableString, LineStream, StreamWriter

from panda3d_gemstone.engine import runtime, prc

from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.runnable import Runnable
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.exceptions import raise_not_implemented

from panda3d_gemstone.logging.utilities import *

class NotifyHandler(StreamHandler):
    """
    Custom StreamHandler for pipeing logging module based messages
    to a Panda3D notifier category
    """

    def __init__(self, name: str = 'python'):
        super().__init__()
        self.__notify = get_notify_category(name)

    def emite(self, record: object) -> None:
        """
        Processes the incoming record from the logging module
        """

        message_parts = self.format(record).split('||')
        level = message_parts[0].lower()
        message = message_parts[1]
        
        if hasattr(self.__notify, level):
            func = getattr(self.__notify, level)
        else:
            func = self.__notify.info
        
        func.info(message)

class ApplicationLogger(Singleton, InternalObject):
    """
    Singleton for managing the applications logging outputs
    """

    def __init__(self):
        Singleton.__init__(self)
        InternalObject.__init__(self)
        self.__stream = None
        self.__setup()

    def __setup(self) -> None:
        """
        Performs setup operations on the application logger
        """

        self.__stream = LineStream()
        Notify.ptr().setOstreamPtr(self.__stream, False)
        Notifier.streamWriter = StreamWriter(Notify.out(), True)
        self.__initialize_logging_module()

    def destroy(self) -> None:
        """
        Performs shutdown operations on the application logger
        """

        Singleton.destroy(self)
        InternalObject.destroy(self)
        
        if self.is_activated():
            self.deactivate()

        for handler_name in self.__handlers:
            self.remove_handler(handler_name)

    def __initialize_logging_module(self) -> None:
        """
        Initializes the Python logging module to pipe through the Panda3D notifier
        """
        level_map = {
            'spam': logging.DEBUG,
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARN,
            'error': logging.ERROR
        }

        level = level_map.get(ConfigVariableString('notify-level-python', '').value, logging.INFO)
        formatter = '%(levelname)s||%(message)s'
        logging.basicConfig(format=formatter, level=level, handlers=[NotifyHandler()])   


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
from panda3d_gemstone.engine import runtime

def event_handler(*args):
    """
    Registers a message listener event handler
    """

    def decorate(f: object) -> None:
        """
        Custom decorator for attaching the __messenger__ attribute
        to the class function for use by the MessengeListener's startup 
        initializer
        """

        setattr(f, '__messenger__', args)

    return decorate

class MessageListener(object):
    """
    Base class for all objects that listen for Messenger events. Accepts manual
    registering of events or automatically via the @event_handler decorator
    """

    def __init__(self, initialize_event_attributes: bool = True):
        if initialize_event_attributes:
            self.__initialize_events()

    def destroy(self) -> None:
        """
        Destroys the MessengerListener object
        """

        self.ignore_all()

    def __initialize_events(self) -> None:
        """
        Initializes the decorator set event handlers
        """

        event_methods = [(name, obj) for name, obj in inspect.getmembers(self) if inspect.ismethod(obj)]
        for name, method in event_methods:
            event = getattr(method, '__messenger__', None)
            if not event:
                continue

            self.accept(event[0], method, *events[1:])

    def accept(self, event: str, method: object, extra_args: list = []) -> object:
        """
        """

        return runtime.messenger.accept(event, self, method, extra_args, 1)

    def accept_once(self, event:str , method: object, extra_args: list = []) -> object:
        """
        """

        return runtime.messenger.accept(event, self, method, extra_args, 0)

    def ignore(self, event: str) -> object:
        """
        """

        return runtime.messenger.ignore(event, self)

    def ignore_all(self) -> object:
        """
        """

        return runtime.messenger.ignoreAll(self)

    def is_accepting(self, event: str) -> object:
        """
        """

        return runtime.messenger.isAccepting(event, self)

    def get_all_accepting(self) -> object:
        """
        """
        
        return runtime.messenger.getAllAccepting(self)

    def is_ignoring(self, event: str) -> object:
        """
        """

        return runtime.messenger.isIgnoring(event, self)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys

from panda3d_gemstone.io.file_system import path_exists
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.service import Service, Command, ServiceManager
from panda3d_gemstone.framework.utilities import save_screenshot, start_background_thread
from panda3d_gemstone.framework.utilities import print_refcounts, print_unreachable_garbage
from panda3d_gemstone.framework.resource import ResourceManager

from panda3d_gemstone.engine.performance import toggle_profiling
from panda3d_gemstone.engine.showbase import ApplicationShowBase
from panda3d_gemstone.engine import runtime
from panda3d_gemstone.engine.http import HTTPManager
from panda3d_gemstone.engine import prc

from panda3d_gemstone.logging import utilities as _logging
from panda3d_gemstone.logging import handler as _logging_handler

from panda3d_gemstone.application.version import ApplicationVersion
from panda3d_gemstone.application import constants
from panda3d_gemstone.application.events import log_event

from panda3d.core import ConfigVariableString, WindowProperties, AntialiasAttrib

def get_home_directory() -> str:
    """
    Returns the applications current home directory
    """

    result = None
    if sys.platform == 'win32':
        result = '.%s' % os.sep
    elif sys.platform == 'darwin':
        if 'HOME' in os.environ:
            result = '%s%s' % (os.environ['HOME'], os.sep)
    else:
        result = '.%s' % os.sep

    return result

class EngineApplication(Service):
    """
    Base class for all Gemstone applications
    """

    def __init__(self):
        super().__init__()

        if runtime.is_interactive():
            print('WARNING: Running from interactive prompt. Complete compatibility is not guaranteed')

    def get_application_version(self) -> str:
        """
        Returns the application version.
        """

        return str(ApplicationVersion.instantiate_singleton())

    def get_application_name(self) -> str:
        """
        Returns the application's name. Intended to be 
        overridden by child objects
        """

        raise NotImplementedError('%s does not implement get_application_name' % self.__class__.__name__)

    def __has_panda_notify(self) -> bool:
        """
        Returns true if the object has a Panda3D notifier
        object
        """

        notify = 'notify'
        found = False
        has_attr = hasattr(self, notify)
        attr = getattr(self, notify)
        if has_attr and attr != None:
            found = not callable(attr)
        
        return found

    def __log_info(self, message: str) -> None:
        """
        Logs an info message to console if notify is available
        otherwise logs through print
        """

        has_notify = self.__has_panda_notify()
        if not has_notify:
            print(message)
        else:
            self.notify.info(message)

    def setup(self) -> None:
        """
        Performs setup operations on the application instance
        """

        self.activate()

    def destroy(self) -> None:
        """
        Performs destruction operations on the application instance
        """

        self.__log_info('Shutting down %s...' % self.get_application_name())
        self.deactivate()

    def quit(self, code: int = 0) -> None:
        """
        Exits the application with the requested code
        """

        print('Exiting with code: %d' % code)
        sys.exit(code)

class GemstoneApplication(EngineApplication, InternalObject):
    """
    Base class for all GemStone Panda3D Applications
    """

    PRC_FILES = []
    CONTROL_MAPPING = ''
    SHOWBASE_CLS = ApplicationShowBase

    def __init__(self, window_type: str = None, showbase_cls: object = None):
        EngineApplication.__init__(self)
        InternalObject.__init__(self)
        self.set_skip_export(prc.get_prc_bool('gs-skip-export', False))
        if runtime.has_application():
            raise Exception('Attempting to spawn more then one Gemstone Panda3D application instance')
        runtime.application = self
        self.load_prc_files(self.PRC_FILES)

        self.__base = None
        self.__http = None
        self.__showbase_cls = showbase_cls or self.SHOWBASE_CLS
        self.__window_type = window_type

        self.before_panda3d_setup()
        self.__setup_panda3d()
        self.post_panda3d_setup()

        if prc.get_prc_bool('gs-background-thread', True):
            start_background_thread()

        self.__load_control_mapping(self.CONTROL_MAPPING or prc.get_prc_string('gs-control-map', ''))

        self.add_command(Command('save screenshot', save_screenshot))
        self.add_command(Command('quit', self.quit))

    @property
    def base(self) -> ApplicationShowBase:
        """
        Applications showbase instance
        """

        return self.__base

    @property
    def http(self) -> HTTPManager:
        """
        Applications HTTPManager instance
        """

        return self.__http

    @property
    def window(self) -> object:
        """
        Applications main window instance
        """

        if not self.__base:
            return None

        return self.__base.win

    def load_prc_files(self, files: list, optional: bool = False) -> None:
        """
        Attempts to load a list of prc files
        """

        if not len(files):
            return

        for prc_file in files:
            prc.load_prc_file(prc_file, optional)

    def before_panda3d_setup(self) -> None:
        """
        Called before setting up the Panda3D game engine
        """

        prc.load_prc_file_data('gemstone', constants.GEMSTONE_PRC)

    def __setup_panda3d(self) -> None:
        """
        Performs Panda3d engine setup operations
        """

        # Initialize showbase
        self.notify.info('Setting up engine...')
        assert self.__showbase_cls != None
        self.__showbase_cls.notify = _logging.get_notify_category('showbase')
        self.__base = self.__showbase_cls(windowType=self.__window_type)
        self.__base.set_exit_callback(self.destroy)

        runtime.window = self.__base.win
        runtime.task_mgr = self.__base.task_mgr
        runtime.camera = self.__base.camera
        self.setup_window()

    def post_panda3d_setup(self) -> None:
        """
        Called after setting up the Panda3D game engine
        """

        # Replace IO access
        self.notify.info('Setting up environment...')
        from panda3d_gemstone.io import file_system
        file_system.switch_file_functions_to_vfs()

        # Setup singletons
        self.__http = HTTPManager.instantiate_singleton()
        runtime.http = self.__http

        self._logger = _logging_handler.ApplicationLogger.get_singleton()
        if self._logger:
            runtime.logger = self._logger

    def setup(self) -> None:
        """
        Performs setup operations on the application instance
        """

        EngineApplication.setup(self)

        # Log the application and version      
        self.notify.info('Starting (%s) %s - %s' % (
            self.__class__.__name__,
            self.get_application_name(),
            self.get_application_version()))

    def destroy(self) -> None:
        """
        Performs destruction operations on the application instance
        """

        EngineApplication.destroy(self)
        InternalObject.destroy(self)

    def run(self) -> None:
        """
        Runs the Geomstone application
        """

        self.base.run()

    def setup_window(self) -> None:
        """
        Performs window setup operations
        """

        self.notify.info('Performing window setup.')
        window_title = self.get_window_label()
        self.__base.set_window_title(window_title)

    def get_window_label(self) -> str:
        """
        Returns the applications window title label
        """

        return '%s - %s' % (
            self.get_application_name(),
            self.get_application_version())

    def __load_control_mapping(self, filename: str) -> None:
        """
        Loads a control map into the service manager
        """

        if not filename:
            return

        ServiceManager.get_singleton().load(filename)

    def set_skip_export(self, state: bool) -> None:
        """
        Sets the resource manager's skip export state
        """

        ResourceManager.get_singleton().set_skip_export(state)

    def get_application_name(self) -> str:
        """
        Returns the application's name. 
        """

        return ConfigVariableString('app-name', 'Gemstone Application').value
 
class HeadlessGemstoneApplication(GemstoneApplication):
    """
    Headless variant of the Panda3D Gemstone framework application
    """

    def __init__(self, *args, **kwargs):
        kwargs['window_type'] = 'none'
        super().__init__(*args, **kwargs)

class ServerGemstoneApplication(HeadlessGemstoneApplication):
    """
    Variant of the headless Panda3D Gemstone framework application
    for use in server applications
    """

    def get_public_ip(self, callback: object) -> None:
        """
        Returns the server application's public ip address
        """

        assert callback != None
        assert callable(callback)

        def wrapper(data):
            """
            Wrapper to break out the callback data
            """

            ip = data.geT('ip', 'unknown')
            callback(ip)

        url = 'https://api.ipify.org/?format=json'
        self.http.perform_json_get_Request(url, callback=wrapper)



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import io
import tokenize

from panda3d.core import Vec2, Vec3, Vec4
from panda3d.core import Point2, Point3, Point4

from panda3d_gemstone.logging.utilities import get_notify_category

__pcast_notify = get_notify_category('pcast')

def __cast_scalar(token: object, name: str, src: object, scalar_types: object) -> object:
    """
    """

    if token[1].lower() == name:
        token = next(src)
        if token[1] == '(':
            out = []
            token = next(src)
            while token[1] != ')':
                out.append(__cast_number(token, src))
                token = next(src)
                if token[1] == ',':
                    token = next(src)

            if len(out) in scalar_types:
                return scalar_types[len(out)](*out)
            else:
                raise ValueError('Malformed scalar type (wrong number of elements)')
        else:
            raise SyntaxError('Malformed expression %s' % str(token))

def __cast_vec(token: object, src: object) -> object:
    """
    """
    
    return __cast_scalar(token, 'vec', src, {
        2: Vec2,
        3: Vec3,
        4: Vec4
    })

def __cast_point(token: object, src: object) -> object:
    """
    """

    return __cast_scalar(token, 'point', src, {
        2: Point2,
        3: Point3,
        4: Point4
    })

def __cast_dict(token: object, src: object) -> object:
    """
    """

    if token[1] == '{':
        out = {}
        token = next(src)
        while token[1] != '}':
            key = __cast(token, src)
            token = next(src)
            if token[1] != ':':
                raise SyntaxError('malformed dictionary')

            value = __cast(next(src), src)
            out[key] = value
            token = next(src)
            if token[1] == ',':
                token = next(src)

        return out

def __cast_sequence(token: object, src: object, start_token: str, end_token: str, seperator: str) -> object:
    """
    """

    if token[1] == start_token:
        out = []
        token = next(src)
        while token[1] != end_token:
            out.append(__cast(token, src))
            token = next(src)
            if token[1] == end_yoken:
                continue
            if token[1] != seperator:
                raise SyntaxError('Malformed sequence')
            token = next(src)

        return out

def __castList(token: object, src: object) -> object:
    return __cast_sequence(token, src, '[', ']', ',')

def __cast_tuple(token: object, src: object) -> object:
    """
    """
    out = __cast_sequence(token, src, '(', ')', ',')
    if out is not None:
        out = tuple(out)
    
    return out

def __castString(token, src):
    if token[0] == tokenize.STRING:
        return str(token[1][1:-1])

def __cast_Name(token: object, src: object):
    """
    """

    nameToCasterMapping = {
        'vec': __castVec,
        'point': __castPoint,
        'false': lambda token, src: False,
        'true': lambda token, src: True,
        'none': lambda token, src: None
    }

    if token[0] == tokenize.NAME:
        name = token[1]
        if name.lower() in name_to_caster_mapping:
            return name_to_caster_mapping[name.lower()](token, src)
        else:
            return name

def __cast_number(token: object, src: object) -> int:
    """
    """

    sign = 1
    if token[0] == tokenize.OP and token[1] == '+':
        sign = 1
        token = next(src)
    elif token[0] == tokenize.OP and token[1] == '-':
        sign = -1
        token = next(src)
    if token[0] == tokenize.NUMBER:
        try:
            return sign * int(token[1], 0)
        except ValueError:
            return sign * float(token[1])

def __cast(token: object, src: object) -> object:
    """
    """

    out = __cast_dict(token, src)
    if out is not None:
        return out

    out = __cast_kist(token, src)
    if out is not None:
        return out

    out = __cast_tuple(token, src)
    if out is not None:
        return out

    out = __cast_string(token, src)
    if out is not None:
        return out

    out = __cast_number(token, src)
    if out is not None:
        return out

    out = __cast_name(token, src)

    return out

def cast(txt: str) -> object:
    if type(txt) in [str, str]:
        txt = txt.strip()
        if txt:
            try:
                src = io.StringIO(txt).readline
                src = tokenize.generate_tokens(src)
                out = __cast(next(src), src)
                for token in src:
                    if token[0] not in (tokenize.NEWLINE, tokenize.ENDMARKER):
                        raise SyntaxError('Malformed expression %s' % str(token))

                return out
            except ValueError as msg:
                __pcast_notify.error('Value Error (pcast): %s', msg)
            except SyntaxError as msg:
                __pcast_notify.error('Syntax Error (pcast): %s', msg)
            except StopIteration:
                raise SyntaxError('Malformed expression')
            except tokenize.TokenError as msg:
                __pcast_notify.error('Syntax Error (pcast): %s', msg)

    return None

def seralize(data: object) -> str:
    """
    """

    replace_mapping = {
        'Vec2': 'Vec',
        'Vec3': 'Vec',
        'Vec4': 'Vec',
        'Point2': 'Point',
        'Point3': 'Point',
        'Point4': 'Point'
    }

    txt = str(data)
    for key, value in list(replace_mapping.items()):
        txt = txt.replace(key, value)

    return txt



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
import base64
import traceback
from collections import defaultdict

from panda3d_gemstone.engine import runtime, prc
from panda3d_gemstone.framework.internal_object import InternalObject

class LauncherExitCodes(object):
    """
    Contains all possible exit codes for the 
    Geomstone launcher
    """

    SUCCESS = 0
    FAILED_STARTUP_CHECK = 1
    INTERNAL_ERROR = 2

class GemstoneLauncher(InternalObject):
    """
    Base class for all Geomstone launchers
    """

    APPLICATION_CLS = None

    def __init__(self, application_cls: object = None):
        self._application_cls = application_cls or self.APPLICATION_CLS
        assert self._application_cls != None
        super().__init__()
        self.exit_code = 0
        runtime.launcher = self

        self.__application = None
        self._launch_arguments = defaultdict(list)
        self.__parse_command_line()

    @property
    def application(self) -> object:
        """
        Launchers current application instance
        """

        return self.get_application()

    def get_application(self) -> object:
        """
        Returns the launcher's current application instance
        """

        return self.__application

    def get_environment_variable(self, key: str, default: str) -> str:
        """
        Returns the environment variable defined at the time of launch if present.
        Otherwise returns the value of default
        """

        return os.environ.get(key, default)

    def has_launch_arg(self, key: str) -> bool:
        """
        Returns true if the launch argument key was present at the time
        of launch via command line
        """

        return self.get_launch_arg(key, None) != None

    def get_launch_arg(self, key: str, default: str) -> str:
        """
        Retrieves a command line launch argument if present.
        Otherwise returns default
        """

        value = self._launch_arguments.get(key, default)
        if isinstance(value, list):
            if len(value) > 0:
                return value[0]

        return value

    def get_encoded_launch_arg(self, key: str, default: str, bypass: bool=True) -> str:
        """
        Retrieves an encoded command line launch argument if present.
        Otherwise returns default
        """

        value = self.get_launch_arg(key, default)
        result = None
        try:
            decoded = base64.b64decode(value)
            result = decoded.decode('utf-8')
        except:
            pass

        if bypass and runtime.is_developer_build():
            result = value

        return result

    def set_development_launch(self) -> None:
        """
        Flags the current application launch as a development 
        launch. Must be performed before starting the application
        """

        if runtime.has_base():
            self.notify.warning('Failed to set as development launch. Application already started')
            return 

        self.notify.warning('Launching as a development build of the Gemstone application (%s)' % self._application_cls.__name__)
        prc.load_prc_file_data('want-dev #t')
        runtime.dev = True

    def __parse_command_line(self) -> None:
        """
        Parses the output of the command line arguments
        """

        self.notify.debug('Parsing command line arguments')
        for k, v in ((k.lstrip('-'), v) for k,v in (a.split('=') for a in sys.argv[1:])):
            self._launch_arguments[k].append(v)

    def attempt_start(self) -> None:
        """
        Attempts to start the Geomstone application
        """

        # Verify we can start
        if not self.can_start():
            self.exit_code = LauncherExitCodes.FAILED_STARTUP_CHECK
            return
        
        self.launch()

    def can_start(self) -> bool:
        """
        Returns true when the launcher can start. Intended to be overriden by 
        a child launcher object
        """

        raise NotImplementedError('%s does not implement can_start!' % self.__class__.__name__)

    def launch(self) -> None:
        """
        Launches the gemstone application
        """

        if runtime.has_application():
            self.notify.warning('Failed to launch application. An application is already running!')
            return

        try:
            self.__application = self._application_cls()
            self.__application.setup()
            self.__application.run()
        except SystemExit:
            pass
        except Exception as e:
            self.exit_code = LauncherExitCodes.INTERNAL_ERROR
            exe_cls, exe, tb = sys.exc_info()
            print('An unexpected error has occured. %s (%s)' % (str(exe), exe_cls.__name__))
            if runtime.is_developer_build():
                print(traceback.format_exc())
            
            self.handle_application_error(exe_cls, exe, traceback.format_exc())

    def handle_application_error(self, error_cls: object, error: Exception, traceback: object) -> None:
        """
        Called when the application encounters an error for custom
        launcher processing
        """

    def start(self) -> int:
        """
        Starts the application launch process
        """

        self.attempt_start()
        self.destroy()

        return self.exit_code

class DeveloperLauncher(GemstoneLauncher):
    """
    Basic developer launcher that will always launch immediately and permits hot reloads
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.notify.setInfo(True)
        self.set_development_launch()

    def can_start(self) -> bool:
        """
        Returns true when the launcher can start. Intended to be overriden by 
        a child launcher object
        """

        return True
    
def application_main(launcher_cls: object) -> int:
    """
    Handler to process the startup of a Gemstone application
    """

    assert launcher_cls != None
    launcher = launcher_cls()
    launcher.start()

    return launcher.exit_code



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import os
import csv
import logging
import string
import codecs

from panda3d_gemstone.framework.utilities import get_snake_case
from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.io.file_system import get_matching_files
from panda3d_gemstone.engine.prc import get_prc_string, get_prc_bool

def utf_8_encoder(unicode_csv_data: object):
    """
    """

    for line in unicode_csv_data:
        yield line.encode('utf-8')

def unicode_csv_reader(unicode_csv_data: object, dialect: object = csv.excel, **kwargs):
    """
    """

    csv_reader = csv.reader(unicode_csv_data, dialect=dialect, **kwargs)
    for row in csv_reader:
        yield [ cell for cell in row ]

class ApplicationLocalizer(Singleton, InternalObject):
    """
    """

    @staticmethod
    def get_backup_text():
        """
        """

        return ''

    def __init__(self, path: str):
        Singleton.__init__(self)
        InternalObject.__init__(self)
        self.__active_language = None
        self.__encoding = get_prc_string('gs-loc-encoding', 'utf-8')
        self.__delimiter = get_prc_string('gs-loc-delimiter', ';')
        self.__warnings = get_prc_bool('gs-loc-warnings', True)
        self.__resources = {}
        self.__load_all_files(path)

    def __load_all_files(self, root: str) -> None:
        """
        """

        self.notify.info('Reading localization files from: %s' % root)
        files = get_matching_files(root, '*.csv')
        if not len(files):
            self.notify.warning('Failed to setup localization. No files found')

        for csv_file in files:
            name = os.path.splitext(os.path.basename(csv_file))[0]
            self.notify.debug('Reading localization file: %s' % name)

            try:
                lang, topic = name.split('_')
            except ValueError:
                self.notify.error('Failed to load localizer csv. %s has broken formatting. (expected <lang_topic.csv)' % name)

            self.__add_topic(id, topic)

            fh = codecs.open(csv_file, 'r', encoding=self.__encoding)
            reader = unicode_csv_reader(fh, delimiter=self.__delimiter, quotechar='"', skipinitialspace=True)
            by_id = {}
            by_name = {}
            last_succesfull = ['', '']

            try: 
                for line_nr, line in enumerate(reader):
                    if len(line) == 3:
                        if line[0].isdigit():
                            by_id[int(line[0])] = line[2]
                        by_name[line[1]] = line[2]
                        last_succesfull = [line[0], line[1]]
                    else:
                        self.notify.info('Last succesful read: %s %s.' % (last_succesfull[0], last_succesfull[1]))
                        self.notify.error('Failed to read file "%s.csv" in line #%s' % (name, line_nr + 1))
            except Exception as e:
                self.notify.error('Failed to read file "%s.csv. Reason: %s' % (name, e))

            fh.close()
            if lang not in self.__resources:
                self.__resources[lang] = {}
            self.__resources[lang][topic] = (by_id, by_name)
            
            if not self.__active_language:
                self.__active_language = lang

    def __add_topic(self, id: object, topic: object) -> None:
        """
        """

        setattr(self, self.__build_get_name(topic), lambda id, topic = topic, **kw: self.__get(topic=topic, id=id, **kw))
        setattr(ApplicationLocalizer, self.__build_get_name(topic), staticmethod(lambda id, topic = topic, **kw: ApplicationLocalizer.__get(ApplicationLocalizer.get_singleton(), topic=topic, id=id, **kw)))
        setattr(self, self.__build_has_name(topic), lambda id, topic = topic: self.__has(topic=topic, id=id))
        setattr(ApplicationLocalizer, self.__build_has_name(topic), staticmethod(lambda id, topic = topic: ApplicationLocalizer.__has(ApplicationLocalizer.get_singleton(), topic=topic, id=id)))

    def destroy(self) -> None:
        """
        Called on localizer destruction
        """

    def __build_method_name(self, action: str, topic: str) -> str:
        """
        """

        return '%s_%s' % (action, get_snake_case(topic))

    def __build_get_name(self, topic: str) -> str:
        """
        """

        return self.__build_method_name('get', topic)

    def __build_has_name(self, topic: str) -> str:
        """
        """

        return self.__build_method_name('has', topic)

    def __attempt_delete_attr(self, space: object, attr: str) -> None:
        """
        """

        try:
            delattr(space, attr)
        except AttributeError:
            pass

    def __del_attr(self, attr: str) -> None:
        """
        """

        self.__attempt_delete_attr(self, attr)
        self.__attempt_delete_attr(self.__class__, attr)
        
    def set_encoding(self, encoding: str) -> None:
        """
        """

        self.__encoding = encoding

    def get_encoding(self) -> str:
        """
        """

        return self.__encoding

    def set_active_language(self, lang: str) -> None:
        """
        """

        if lang in self.__resources:
            self.notify.info('Setting active language: %s' % lang)
            self.__active_language = lang
        else:
            self.notify.warning('Failed to set active language. Unknown language: %s' % (lang))

    def get_active_language(self) -> str:
        """
        """

        return self.__active_language
    
    def get_all_languages(self) -> list:
        """
        """

        return list(self.__resources.keys())

    def __get_keys(self, lang: str, topic: str) -> tuple:
        """
        """

        if not len(self.__resources):
            self.notify.warning('Failed to retrieve keys. No language data loaded')
            return ([], [])

        if lang in self.__resources:
            language_set = self.__resources[lang]
            if topic in language_set:
                topic_set = language_set[topic]
                return (topic_set[0], topic_set[1])
            else:
                self.notify.warning('Failed to retrieve keys. Unknown topic "%s"' % topic)
        elif self.__warnings:
            self.notify.warning('Failed to retrieve keys. Unknown language "%s"' % lang) 

        return ([], [])

    def get_id_keys(self, lang: str, topic: str) -> list:
        """
        """

        by_id, by_name = self.__get_keys(lang, topic)
        return list(by_id.keys())

    def get_name_keys(self, lang: str, topic: str) -> list:
        """
        """

        by_id, by_name = self.__get_keys(lang, topic)
        return list(by_name.keys())

    def get_keys(self, lang: str, topic: str) -> list:
        """
        """

        by_id, by_name = self.__get_keys(lang, topic)
        return list(by_id.keys()) + list(by_name.keys())

    def __get_template(self, topic: str, id: object) -> object:
        """
        """

        return self.__get_language_template(topic, id, self.__active_language)

    def __get_language_template(self, topic: str, id: object, lang: str) -> object:
        """
        """

        if lang is not None:
            lang = self.__active_language

        result = None
        by_id, by_name = self.__get_keys(lang, topic)

        if id in by_id:
            result = by_id[id]
        
        if id in by_name:
            result = by_name[id]

        return result

    def __decode_special(self, m: object) -> object:
        """
        """

        x = m.group(0)[1:]
        return chr(int(x))

    def __get(self, topic: object, id: object, **kw) -> None:
        """
        """

        result = ApplicationLocalizer.get_backup_text()
        template = self.__get_language_template(topic, id, kw.get('client_language', self.__active_language))
        if template is not None:
            if self.__encoding is None:
                result = template
            else:
                result = template.encode(self.__encoding)

            if kw:
                try:
                    result = string.Template(result).safe_substitute(kw)
                except Exception as e:
                    self.notify.error('Encoding failed: %s (id: %s, kw: %s)' % (id, repr(kw)))

        if not isinstance(result, str):
            result = result.decode(self.__encoding)

        result = result.replace('\\n', '\n')
        return result

    def __has(self, topic: object, id: object) -> object:
        """
        Returns true if the requested topic exists
        """

        return self.__get_template(topic, id) is not None

    def __getattr__(self, key) -> None:
        """
        Custom attribute handler for warning of unknown localizations
        """

        action = key[:3]
        topic = key[4:].lower()

        if action == 'get':
            self.notify.warning('Unknown localization topic: "%s"' % topic)
            return lambda *args, **kw: str('Localization Error')
        elif action == 'has':
            self.notify.warning('Unknown localization topic: "%s"' % topic)
            return lambda *args, **kw: False

        raise AttributeError('%s has no attribute "%s"' % (
            self.__class__.__name__, key))


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from PyQt5.QtGui import QIcon, QPixmap

from panda3d_gemstone.framework.internal_object import InternalObject

class StyleManager(InternalObject):
    """
    Styling manager for the PyQt5 editor application
    """

    def __init__(self, application):
        super().__init__()
        self._application = application
        self._icon_path = 'assets/icons/black'

    def get_icon(self, path):
        """
        Retrieves the icon related to the requested style
        """

        final_path = '%s/%s' % (self._icon_path, path)
        return QIcon(final_path)

    def get_pixmap(self, path):
        """
        Retrieves the pixmap related to the requested style
        """

        final_path = '%s/%s' % (self._icon_path, path)
        return QPixmap(final_path)

    def set_dark_theme(self):
        """
        Attempts to set the editor's dark theme if the required module is available
        """

        has_dark = True
        try:
            import qdarkstyle
        except:
            has_dark = False

        if not has_dark:
            self.notify.warning('Failed to set dark theme; qdarkstyle not installed')
            return

        stylesheet = qdarkstyle.load_stylesheet_pyqt5()
        self._icon_path = 'assets/icons/white'
        self._application.q_application.setStyleSheet(stylesheet)
        self.notify.info('Configured dark theme')    


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from direct.interval.IntervalGlobal import Sequence, Wait, Func

from panda3d_gemstone.framework.internal_object import InternalObject

class CooldownCounter(InternalObject):
    """
    """

    def __init__(self, initial_value: object = 0.0, callback: object = None, *callback_args):
        self.initial_value = initial_value
        self.callback = callback
        self.callback_args = callback_args
        self.interval = None
        self.is_stopping = False

    @property
    def time(self) -> float:
        """
        Returns this counter's time as a property
        """

        return self.get_time()

    def get_time(self) -> float:
        """
        Returns this counters interval time
        """

        if self.interval:
            return self.interval.getT()

        return 0.0

    def set_speed(self, speed: float) -> None:
        """
        Sets this counter's current speed
        """

        self.speed = speed
    
    def get_speed(self) -> float:
        """
        Returns this counter's current speed
        """

        return self.speed

    def is_running(self) -> bool:
        """
        Returns true if the counter is currently running
        """

        if self.interval:
            return self.interval.isPlaying()

        return False

    def is_finished(self) -> bool:
        """
        Returns true if the counter is currently finished
        """

        if self.interval:
            return self.interval.isStopped() and not self.interval.isPlaying()
        return True

    def is_paused(self) -> bool:
        """
        Returns true if the counter is currently paused
        """

        if self.interval:
            return not self.interval.isPlaying() and not self.interval.isStopped()
        return False

    def __begin_interval(self, interval_func_name: str, initial_value: object, callback: object, *callback_args) -> None:
        """
        """

        # Stop any existing intervals
        self.stop(True)

        if initial_value is not None:
            self.initial_value = initial_value
        
        if callback is not None:
            self.callback = callback
            self.callback_args = callback_args
    
        if self.callback:
            self.interval = Sequence(Wait(self.initial_value), Func(self.callback, *callback_args))
        else:
            self.interval = Sequence(Wait(self.initial_value))

        interval_func = getattr(self.interval, interval_func_name, None)
        if callable(interval_func):
            interval_func()
        else:
            self.notify.error('Failed to start interval. Invalid interval function: %s' % interval_func_name)

    def start(self, initial_value: object, callback: object = None, *callback_args) -> None:
        """
        Starts the counter object interval
        """

        self.__begin_interval('start', initial_value, callback, *callback_args)

    def loop(self, initial_value: object, callback: object = None, *callback_args) -> None:
        """
        Loops the counter object interval
        """

        self.__begin_interval('loop', initial_value, callback, *callback_args)

    def stop(self, ignore_callback: bool = False) -> None:
        """
        Stops the counter object interval
        """

        if self.interval and not self.is_stopping():

            if ignore_callback:
                self.interval.clearIntervals()

            self.is_stopping = True
            self.interval.finish()
            self.interval = None
            self.is_stopping = False

    def pause(self) -> None:
        """
        Pauses the counter object interval
        """

        if self.interval:
            self.interval.pause()

    def resume(self) -> None:
        """
        Resumes the counter object interval
        """

        if self.interval:
            self.interval.resume()


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

from direct.directnotify.DirectNotifyGlobal import directNotify

#from anvil.core import editor_globals
#from anvil.widget import properties, layers, asset_library

from functools import partial
import sys

class MenuBase(object):
    """
    Base class for all window menus
    """

    def __init__(self, window, menu_bar):
        self.window = window
        self.menu_bar = menu_bar
        self.notify = directNotify.newCategory(self.__class__.__name__)
        self.build_menu()

    def build_menu(self):
        """
        Constructs the menu object on the window
        """

class EditorFileMenu(MenuBase):
    """
    File menu for the primary editor window
    """

    def build_menu(self):
        """
        Constructs the menu object on the window
        """

        self.file_menu = self.menu_bar.add_menu('&File')
        self.file_menu.add_action('&New', theme_icon='star.png', tooltip='Creates a new level project')
        self.file_menu.add_action('&Save', theme_icon='save.png', tooltip='Saves current project')
        self.file_menu.add_action('&Settings', theme_icon='wrench.png', tooltip='User settings')
        self.file_menu.add_action('&Quit', theme_icon='exit.png', tooltip='Closes the editor', connect=sys.exit)

class EditorEditMenu(MenuBase):
    """
    Edit menu for the primary editor window
    """

    def build_menu(self):
        """
        Constructs the menu object on the window
        """

        self.edit_menu = self.menu_bar.add_menu('&Edit')

class EditorViewMenu(MenuBase):
    """
    View menu for the primary editor window
    """

    def build_menu(self):
        """
        Constructs the menu object on the window
        """

        self.view_menu = self.menu_bar.add_menu('&View')
        #self.view_menu.add_action(
        #    '&Properties Panel', 
        #    tooltip='Opens/Closes the properties panel', 
        #    checkable=True, 
        #    connect=partial(
        #        self.toggle_panel, 
        #        properties.QPropertiesWidget))
        #self.view_menu.add_action(
        #    '&Layers Panel', 
        #    tooltip='Opens/Closes the layers panel', 
        #    checkable=True, 
        #    connect=partial(
        #        self.toggle_panel, 
        #        layers.QLayersWidget))
        #self.view_menu.add_action(
        #    '&Asset Library', 
        #    tooltip='Opens/Closes the asset library', 
        #    checkable=True, 
        #    connect=partial(
        #        self.toggle_panel, 
        #        asset_library.QObjectLibraryWidget,
        #        dock=Qt.LeftDockWidgetArea))

    def toggle_panel(self, *args, **kwargs):
        """
        Toggles the requested panel class
        """

        self.window.toggle_panel(*args, **kwargs)

class EditorHelpMenu(MenuBase):
    """
    Help menu for the primary editor window
    """

    def build_menu(self):
        """
        Constructs the menu object on the window
        """

        self.help_menu = self.menu_bar.add_menu('&Help')
        self.help_menu.add_action('&Docs', theme_icon='question.png', tooltip='Editor documentation', connect=self.open_docs)
        self.help_menu.add_action('&About', theme_icon='information.png', tooltip='About this application', connect=self.open_about)
        #self.help_menu.add_action('&Qt', theme_icon='information.png', tooltip="About this application's Qt version", connect=editor_globals.app.aboutQt)

    def open_docs(self):
        """
        Opens the editor's documentation page
        """

    def open_about(self):
        """
        Opens the editor's about dialog box
        """

        self.notify.info('Opening about dialog box.')
        about_message = """
        %(app_name)s - Gemstone Editor (%(app_version)s)

        Written by:
        Jordan Maxwell <me@jordan-maxwell.info>

        Copyright %(org_name)s, 2020
        %(org_domain)s
        """

        about_info = {
            'app_name': config.GetString('app-name', 'Anvil Level Editor'),
            'app_version': config.GetString('app-version', '1.0.0'),
            'org_name': config.GetString('org-name', 'Digital Descent, LLC'),
            'org_domain': config.GetString('org-domain', 'https://www.digitaldescent.co')
        }

        about_message = about_message % about_info
        response = QMessageBox.about(self.window, 'About Anvil', about_message)

class EditorDebugMenu(MenuBase):
    """
    Debug menu for the primary editor window
    """

    def build_menu(self):
        """
        Constructs the menu object on the window
        """

        self.debug_menu = self.menu_bar.add_menu('&Debug')
        self.debug_menu.add_action(
            '&Print Scene', 
            theme_icon='information.png', 
            tooltip='Prints the entire scene graph to console',
            connect=self.print_scene)
        self.debug_menu.add_action(
            '&Analyze', 
            theme_icon='information.png', 
            tooltip='Analyzes the entire scene graph in console',
            connect=self.analyze_scene)
        self.debug_menu.add_action(
            '&Dump Scene', 
            theme_icon='save.png', 
            tooltip='Dumps the entire scene graph to a file',
            connect=self.dump_scene)

    def print_scene(self):
        """
        Prints the current scene to console
        """

        base.render.ls()
        self.window.status_bar.show_temp_message('Scene printed. Check console for results')

    def analyze_scene(self):
        """
        Analyzes the current scene to console
        """

        base.render.analyze()
        self.window.status_bar.show_temp_message('Scene Analyzed. Check console for results')

    def dump_scene(self):
        """
        Dumps the current editor scene to a file
        """

        base.render.writeBamFile('dump.bam')
        self.window.status_bar.show_temp_message('Scene dumped to `dump.bam`')